# %% [markdown]
# # Quickstart (.py)
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/demo/scripts](https://github.com/barneydobson/wsi/blob/main/docs/demo/scripts/quickstart_demo.py)
#
# 1. [Introduction](#introduction)
#
# 2. [Data](#imports-and-forcing-data)
#
# 3. [Create nodes](#create-nodes)
#
# 4. [Create arcs](#create-arcs)
#
# 5. [Create model](#create-model)
#
# 6. [Run model](#run-model)
#
# 7. [What next?](#what-next?)
# %% [markdown]
# ## Introduction
#
# WSIMOD is a Python package that lets you create nodes that represent physical
# things in the water cycle and enables them talk to each other. In this demo,
# we will create some nodes, some arcs (the things that link nodes), and make
# a model that lets us simulate the flow of water through the water cycle.
#
# We will create a simple catchment model that contains a hydrological node,
# a sewer node, a groundwater node, and a river node. Using these nodes, we
# will simulate a mixed urban-rural catchment.
#
# %% [markdown]
# ## Imports and forcing data

# %% [markdown]
# Import packages
# %%

import os

import pandas as pd
from matplotlib import pyplot as plt

from wsimod.core import constants
from wsimod.orchestration.model import Model

# %% [markdown]
# Load input data, in this example we use precipitation, temperature, and
# evapotranspiration (et0).
# %%

# Select the root path for the data folder. Use the appropriate value for your case.
data_folder = os.path.join(os.path.abspath(""), "docs", "demo", "data")


input_fid = os.path.join(data_folder, "processed", "timeseries_data.csv")
input_data = pd.read_csv(input_fid)
input_data.loc[input_data.variable == "precipitation", "value"] *= constants.MM_TO_M
input_data.date = pd.to_datetime(input_data.date)
input_data = input_data.loc[input_data.site == "oxford_land"]
dates = input_data.date.drop_duplicates()
print(input_data.sample(10))

# %% [markdown]
# Input data is stored in dicts, where a key is a variable on a given day.
# %%

land_inputs = input_data.set_index(["variable", "date"]).value.to_dict()

example_date = pd.to_datetime("2009-03-03")

print(land_inputs[("precipitation", example_date)])
print(land_inputs[("et0", example_date)])
print(land_inputs[("temperature", example_date)])

# %% [markdown]
# ## Create nodes
# %% [markdown]
# Nodes can be defined as dictionaries of parameters. Different nodes require
# different parameters, you can see the documentation in the
# [API](./../../../reference) to understand what parameters can be set.
# Although every node should have a type (type_) and a name.
# %%

sewer = {"type_": "Sewer", "capacity": 0.04, "name": "my_sewer"}

surface1 = {
    "type_": "ImperviousSurface",
    "surface": "urban",
    "area": 10,
    "pollutant_load": {"phosphate": 1e-7},
}

surface2 = {
    "type_": "PerviousSurface",
    "surface": "rural",
    "area": 100,
    "depth": 0.5,
    "pollutant_load": {"phosphate": 1e-7},
}

land = {
    "type_": "Land",
    "data_input_dict": land_inputs,
    "surfaces": [surface1, surface2],
    "name": "my_land",
}

gw = {"type_": "Groundwater", "area": 100, "capacity": 100, "name": "my_groundwater"}

node = {"type_": "Node", "name": "my_river"}

waste = {"type_": "Waste", "name": "my_outlet"}

# %% [markdown]
# ## Create arcs
# %% [markdown]
# [Arcs](./../../../reference-arc/#wsimod.arcs.arcs.Arc) can also be
# created as dictionaries, they don't typically need any
# numerical parameters (although there are some exceptions in the
# [case study demo](./../oxford_demo/#Arc-parameters)). Though they do need
# to specify the in_port (where the arc starts), and out_port (where it
# finishes). It's also handy to give each arc a name.
# %%
urban_drainage = {
    "type_": "Arc",
    "in_port": "my_land",
    "out_port": "my_sewer",
    "name": "urban_drainage",
}

percolation = {
    "type_": "Arc",
    "in_port": "my_land",
    "out_port": "my_groundwater",
    "name": "percolation",
}

runoff = {
    "type_": "Arc",
    "in_port": "my_land",
    "out_port": "my_river",
    "name": "runoff",
}

storm_outflow = {
    "type_": "Arc",
    "in_port": "my_sewer",
    "out_port": "my_river",
    "name": "storm_outflow",
}

baseflow = {
    "type_": "Arc",
    "in_port": "my_groundwater",
    "out_port": "my_river",
    "name": "baseflow",
}

catchment_outflow = {
    "type_": "Arc",
    "in_port": "my_river",
    "out_port": "my_outlet",
    "name": "catchment_outflow",
}

# %% [markdown]
# ## Create model
# %% [markdown]
# We can create a [model object](./../../../reference-model/#wsimod.orchestration.model.Model) and add dates
# %%
my_model = Model()
my_model.dates = dates
# %% [markdown]
# Add nodes in a list. The Model object will create the nodes from the dict
# entries.
# %%
my_model.add_nodes([sewer, land, gw, node, waste])

# %% [markdown]
# Add arcs in a list.
# %%
my_model.add_arcs(
    [urban_drainage, percolation, runoff, storm_outflow, baseflow, catchment_outflow]
)
# %% [markdown]
# ## Run model
# %% [markdown]
# Run the model
# %%
flows, _, _, _ = my_model.run()
flows = pd.DataFrame(flows)

# %% [markdown]
# Plot results
# %%
f, axs_ = plt.subplots(3, 2, figsize=(10, 10))

for axs, variable in zip(axs_.T, ["flow", "phosphate"]):
    flows_plot = flows.pivot(index="time", columns="arc", values=variable)
    input_data.pivot(index="date", columns="variable", values="value")[
        ["precipitation"]
    ].plot(ax=axs[0], xlabel="", xticks=[], sharex=True)

    flows_plot[["runoff", "urban_drainage", "percolation"]].plot(
        ax=axs[1], xlabel="", xticks=[], sharex=True
    )
    flows_plot[["catchment_outflow", "storm_outflow", "baseflow"]].plot(ax=axs[2])
axs_[1, 0].set_title("Flows (m3/d)")
axs_[1, 1].set_title("Phosphate (kg/d)")
f.tight_layout()

# %% [markdown]
# ## What next?
# If you are hydrologically minded then you might be interested in a more detailed
# tutorial for our [Land node](./../land_demo). If you want an overview to
# see how many different nodes in WSIMOD work, then the
# [Oxford case study](./../oxford_demo) might be more interesting.



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.catchment import Catchment
from wsimod.nodes.waste import Waste


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_get_flow(self):
        catchment = Catchment(
            name="",
            data_input_dict={
                ("flow", 1): 10,
                ("phosphate", 1): 0.4,
                ("temperature", 1): 10,
            },
        )
        catchment.t = 1
        vq = catchment.get_flow()
        d1 = {"volume": 10, "phosphate": 0.4 * 10, "temperature": 10}
        self.assertDictAlmostEqual(d1, vq)

    def test_get_avail(self):
        catchment = Catchment(
            name="",
            data_input_dict={
                ("flow", 1): 10,
                ("phosphate", 1): 0.4,
                ("temperature", 1): 10,
            },
        )
        waste = Waste(name="")
        arc1 = Arc(in_port=catchment, out_port=waste, name="")
        arc1.vqip_in = {"volume": 2}
        catchment.t = 1
        vq = catchment.get_avail()
        d1 = {"volume": 10 - 2, "phosphate": 0.4 * 8, "temperature": 10}
        self.assertDictAlmostEqual(d1, vq)

    def test_route(self):
        catchment = Catchment(
            name="",
            data_input_dict={
                ("flow", 1): 10,
                ("phosphate", 1): 0.4,
                ("temperature", 1): 10,
            },
        )
        catchment.t = 1

        waste = Waste(name="")
        arc1 = Arc(in_port=catchment, out_port=waste, name="")

        catchment.get_flow()
        d1 = {"volume": 10, "phosphate": 0.4 * 10, "temperature": 10}

        catchment.route()

        self.assertDictAlmostEqual(d1, arc1.vqip_out)

    def test_pull(self):
        catchment = Catchment(
            name="",
            data_input_dict={
                ("flow", 1): 10,
                ("phosphate", 1): 0.4,
                ("temperature", 1): 10,
            },
        )
        waste = Waste(name="")
        arc1 = Arc(in_port=catchment, out_port=waste, name="")
        arc1.vqip_in["volume"] = 2
        catchment.t = 1
        vq1 = catchment.pull_check_abstraction()
        d1 = {"volume": 10 - 2, "phosphate": 0.4 * 8, "temperature": 10}
        self.assertDictAlmostEqual(d1, vq1)

        vq2 = catchment.pull_check_abstraction({"volume": 3})
        d2 = {"volume": 3, "phosphate": 0.4 * 3, "temperature": 10}
        self.assertDictAlmostEqual(d2, vq2)

        vq2 = catchment.pull_set_abstraction({"volume": 3})
        self.assertDictAlmostEqual(d2, vq2)


if __name__ == "__main__":
    unittest.main()



from pytest import mark
from pathlib import Path
import subprocess


def collect_examples() -> list[Path]:
    root = Path.cwd() / "docs" / "demo" / "examples"
    return list(root.glob("**/*.yaml"))


@mark.parametrize("example", collect_examples())
def test_examples(example: Path, tmp_path: Path) -> None:
    result = subprocess.run(
        f"wsimod {str(example)} -o {str(tmp_path)}",
        shell=True,
        check=True,
    )
    assert (tmp_path / "flows.csv").exists()
    assert (tmp_path / "tanks.csv").exists()
    assert (tmp_path / "surfaces.csv").exists()



"""The entry point for the myproject program."""

from argparse import ArgumentParser
from pathlib import Path
from typing import Any, cast

import pandas as pd

from wsimod.orchestration.model import Model
from wsimod.validation import (
    _validate_input_dir,
    _validate_output_dir,
    assign_data_to_settings,
    evaluate_input_file,
    load_data_files,
    validate_io_args,
)


def create_parser() -> ArgumentParser:
    """Create the CLI argument parser."""
    parser = ArgumentParser(prog="WSIMOD")
    parser.add_argument(
        "settings",
        type=Path,
        help="Path to the WSIMOD input file, in YAML format.",
    )
    parser.add_argument(
        "--inputs",
        "-i",
        type=Path,
        help="Base directory for all input files. If present, overwrites value in the"
        " settings file.",
    )
    parser.add_argument(
        "--outputs",
        "-o",
        type=Path,
        help="Base directory for all output files. If present, overwrites value in the"
        " settings file.",
    )

    return parser


def run_model(settings: dict[str, Any], outputs: Path) -> None:
    """Runs the mode with the chosen settings and saves the outputs as csv.

    Args:
        settings (dict[str, Any]): Settings dictionary with loaded data.
        outputs(Path): Directory where to save the outputs.
    """
    model = Model()

    model.dates = cast(pd.Series, settings["dates"]).drop_duplicates()
    model.add_nodes(settings["nodes"])
    model.add_arcs(settings["arcs"])

    flows, tanks, _, surfaces = model.run()

    pd.DataFrame(flows).to_csv(outputs / "flows.csv")
    pd.DataFrame(tanks).to_csv(outputs / "tanks.csv")
    pd.DataFrame(surfaces).to_csv(outputs / "surfaces.csv")


def run_saved_model(settings: str, inputs: Path, outputs: Path) -> None:
    """Runs a previously saved model.

    Args:
        settings (str): The name of the settings file to load.
        inputs (Path): The location of that file, as well as any other input file.
        outputs (Path): Directory where to save the outputs.
    """
    model = Model()
    model.load(inputs, config_name=settings)
    flows, tanks, _, surfaces = model.run()

    pd.DataFrame(flows).to_csv(outputs / "flows.csv")
    pd.DataFrame(tanks).to_csv(outputs / "tanks.csv")
    pd.DataFrame(surfaces).to_csv(outputs / "surfaces.csv")


def run() -> None:
    """Main entry point of the application."""
    args = vars(create_parser().parse_args())
    settings_type = evaluate_input_file(args["settings"])

    if settings_type == "custom":
        settings = validate_io_args(**args)
        inputs = settings.pop("inputs")
        outputs = settings.pop("outputs")
        loaded_data = load_data_files(settings.pop("data", {}), inputs)
        loaded_settings = assign_data_to_settings(settings, loaded_data)
        run_model(loaded_settings, outputs)
    else:
        settings_file: Path = args["settings"]
        inputs = _validate_input_dir(args["inputs"], settings_file.parent)
        outputs = _validate_output_dir(args["outputs"], settings_file.parent)
        run_saved_model(settings_file.name, inputs, outputs)


if __name__ == "__main__":
    run()



from wsimod.nodes.nodes import Node


class CustomNode(Node):
    """A custom node."""

    def __init__(self, name):
        """Initialise the node."""
        super().__init__(name)
        self.custom_attr = 1

    def end_timestep(self):
        self.custom_attr += 1
        super().end_timestep()



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.arcs.arcs import AltQueueArc, Arc, DecayArc, DecayArcAlt, QueueArc
from wsimod.core import constants
from wsimod.nodes.nodes import Node
from wsimod.nodes.storage import Storage
from wsimod.nodes.waste import Waste


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def get_simple_model1(self):
        """

        Returns:

        """
        node1 = Storage(name="1", capacity=20)
        node2 = Waste(name="2")
        node3 = Node(name="3")
        node4 = Node(name="4")
        node5 = Waste(name="5")

        arc1 = Arc(in_port=node1, out_port=node2, name="arc1")

        arc2 = Arc(in_port=node3, out_port=node1, name="arc2")

        arc3 = Arc(in_port=node4, out_port=node1, name="arc3")

        arc4 = Arc(in_port=node1, out_port=node5, preference=0.5, name="arc4")

        return node1, node2, node3, node4, node5, arc1, arc2, arc3, arc4

    def test_arc_mb(self):
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 0, "phosphate": 0, "temperature": 0}
        d3 = {"volume": 20, "phosphate": 0.0002, "temperature": 15}
        d4 = {"volume": 10, "phosphate": 0, "temperature": 0}
        d5 = {"volume": 5, "phosphate": 10, "temperature": 10}
        node1 = Node(name="1")
        node2 = Waste(name="2")

        arc1 = Arc(in_port=node1, out_port=node2, name="arc1", capacity=10)

        reply = arc1.send_push_check()
        self.assertDictAlmostEqual(d4, reply, 16)

        reply = arc1.send_push_check(d5)
        self.assertDictAlmostEqual(d5, reply, 16)

        arc1.send_push_request(d1)

        in_, ds_, out_ = arc1.arc_mass_balance()

        self.assertDictAlmostEqual(d1, in_, 16)
        self.assertDictAlmostEqual(d2, ds_, 16)
        self.assertDictAlmostEqual(d1, out_, 16)

        reply = arc1.send_push_request(d1)
        self.assertDictAlmostEqual(d1, reply, 16)

        reply = arc1.send_push_request(d1, force=True)
        in_, ds_, out_ = arc1.arc_mass_balance()
        self.assertDictAlmostEqual(d3, in_, 16)

    def test_arc_pull(self):
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 15}
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()
        node1.push_set(d1)

        reply = arc1.send_pull_check()
        self.assertDictAlmostEqual(d1, reply, 16)

        reply = arc1.send_pull_check({"volume": 5})
        self.assertDictAlmostEqual(d2, reply, 16)

        reply = arc1.send_pull_request({"volume": 5})
        self.assertDictAlmostEqual(d2, reply, 16)

        reply = arc1.send_pull_request({"volume": 10})
        self.assertDictAlmostEqual(d2, reply, 16)

    def test_enter_arcq(self):
        node1 = Node(name="1")
        node2 = Waste(name="2")

        arc1 = QueueArc(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
        )
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}

        request = arc1.enter_arc({"vqip": d1, "time": 1}, "pull", "hi")
        self.assertEqual(5, request["average_flow"])
        self.assertEqual(5, arc1.flow_in)
        self.assertEqual("pull", request["direction"])
        self.assertEqual("hi", request["tag"])
        self.assertDictAlmostEqual(d1, request["vqip"])
        self.assertDictAlmostEqual(d1, arc1.vqip_in)

    def test_enter_q_arcq(self):
        node1 = Node(name="1")
        node2 = Waste(name="2")

        arc1 = QueueArc(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
        )
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        req = {"vqip": d1, "time": 1}
        arc1.enter_queue(req, direction="pull", tag="hi")

        self.assertEqual(5, arc1.queue[0]["average_flow"])
        self.assertEqual(5, arc1.flow_in)
        self.assertEqual("pull", arc1.queue[0]["direction"])
        self.assertEqual("hi", arc1.queue[0]["tag"])
        self.assertDictAlmostEqual(d1, arc1.queue[0]["vqip"])
        self.assertDictAlmostEqual(d1, arc1.vqip_in)

    def test_queue_arc_push(self):
        node1 = Node(name="1")
        node2 = Waste(name="2")

        arc1 = QueueArc(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
            number_of_timesteps=1,
        )

        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}

        arc1.send_push_request(d1)

        self.assertDictAlmostEqual(d1, arc1.vqip_in)
        self.assertDictAlmostEqual(d1, arc1.queue[0]["vqip"])

        arc1.end_timestep()

        d2 = {"volume": 0, "phosphate": 0, "temperature": 0}
        self.assertDictAlmostEqual(d2, arc1.vqip_in)
        self.assertDictAlmostEqual(d1, arc1.queue[0]["vqip"])

        _ = arc1.update_queue(direction="push")
        self.assertDictAlmostEqual(d1, arc1.vqip_out)

    def test_queue_arc_pull(self):
        node1 = Node(name="1")
        node2 = Storage(name="2", capacity=10, initial_storage=5)

        arc1 = QueueArc(
            in_port=node2,
            out_port=node1,
            name="arc1",
            capacity=10,
            number_of_timesteps=1,
        )

        d1 = {"volume": 2.5}
        d2 = {"volume": 0, "phosphate": 0, "temperature": 0}

        reply = arc1.send_pull_request(d1)

        self.assertDictAlmostEqual(d2, reply)

        d2["volume"] = 2.5

        self.assertDictAlmostEqual(d2, arc1.queue[0]["vqip"])
        self.assertDictAlmostEqual(d2, arc1.vqip_in)

        arc1.end_timestep()

        self.assertDictAlmostEqual(d2, arc1.queue[0]["vqip"])

        _ = arc1.update_queue(direction="pull")
        self.assertDictAlmostEqual(d2, arc1.vqip_out)

    def test_altqueue_arc_sum(self):
        node1 = Node(name="1")
        node2 = Waste(name="2")

        arc1 = AltQueueArc(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
            number_of_timesteps=1,
        )

        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}

        arc1.queue[1] = d1
        arc1.queue[0] = d1

        d2 = {"volume": 20, "phosphate": 0.0002, "temperature": 15}

        self.assertDictAlmostEqual(d2, arc1.alt_queue_arc_sum())

    def test_enter_q_arcalt(self):
        node1 = Node(name="1")
        node2 = Waste(name="2")

        arc1 = AltQueueArc(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
        )
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        req = {"vqip": d1, "time": 1}
        arc1.enter_queue(req, direction="push", tag="hi")

        self.assertEqual(5, arc1.flow_in)
        self.assertDictAlmostEqual(d1, arc1.queue[1])
        self.assertDictAlmostEqual(d1, arc1.vqip_in)

        arc1.end_timestep()

        self.assertEqual(0, arc1.flow_in)
        self.assertDictAlmostEqual(d1, arc1.queue[0])

        d2 = {"volume": 0, "phosphate": 0, "temperature": 0}
        self.assertDictAlmostEqual(d2, arc1.vqip_in)

        arc1.update_queue()
        self.assertDictAlmostEqual(d2, arc1.queue[0])
        self.assertDictAlmostEqual(d1, arc1.vqip_out)

    def test_decay_arc(self):
        node1 = Node(name="1", data_input_dict={("temperature", 1): 10})
        node2 = Waste(name="2")
        node1.t = 1
        arc1 = DecayArc(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
            decays={"phosphate": {"constant": 0.005, "exponent": 1.005}},
            number_of_timesteps=1,
        )
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}

        arc1.send_push_request(d1)
        self.assertDictAlmostEqual(d1, arc1.vqip_in)
        d1["phosphate"] = d1["phosphate"] * (1 - 0.005 * 1.005 ** (10 - 20))
        self.assertDictAlmostEqual(d1, arc1.queue[0]["vqip"])

        arc1.end_timestep()
        d1["phosphate"] = d1["phosphate"] * (1 - 0.005 * 1.005 ** (10 - 20))
        self.assertDictAlmostEqual(d1, arc1.queue[0]["vqip"])

    def test_decay_arc_alt(self):
        node1 = Node(name="1", data_input_dict={("temperature", 1): 10})
        node2 = Waste(name="2")
        node1.t = 1
        arc1 = DecayArcAlt(
            in_port=node1,
            out_port=node2,
            name="arc1",
            capacity=10,
            number_of_timesteps=1,
            decays={"phosphate": {"constant": 0.005, "exponent": 1.005}},
        )
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}

        arc1.send_push_request(d1)
        self.assertDictAlmostEqual(d1, arc1.vqip_in)
        d1["phosphate"] = d1["phosphate"] * (1 - 0.005 * 1.005 ** (10 - 20))
        self.assertDictAlmostEqual(d1, arc1.queue[1])

        arc1.end_timestep()
        d1["phosphate"] = d1["phosphate"] * (1 - 0.005 * 1.005 ** (10 - 20))
        self.assertDictAlmostEqual(d1, arc1.queue[0])

    def test_overrides_simple_arc(self):
        model = self.get_simple_model1()
        arc1 = model[5]
        arc1.apply_overrides({"capacity": 3.1, "preference": 0.65})
        self.assertEqual(3.1, arc1.capacity)
        self.assertEqual(0.65, arc1.preference)

        # Test incorrect label
        overrides = {"capacity": 3.1, "preferenceasd": 0.65}
        arc1.apply_overrides(overrides)
        self.assertEqual({"preferenceasd": 0.65}, overrides)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.nodes import Node
from wsimod.nodes.storage import Storage
from wsimod.nodes.waste import Waste
import os
import pandas as pd


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def get_simple_model1(self):
        """

        Returns:

        """
        node1 = Storage(name="1", capacity=20)
        node2 = Waste(name="2")
        node3 = Node(name="3")
        node4 = Node(name="4")
        node5 = Waste(name="5")

        arc1 = Arc(in_port=node1, out_port=node2, name="arc1")

        arc2 = Arc(in_port=node3, out_port=node1, name="arc2")

        arc3 = Arc(in_port=node4, out_port=node1, name="arc3")

        arc4 = Arc(in_port=node1, out_port=node5, preference=0.5, name="arc4")

        return node1, node2, node3, node4, node5, arc1, arc2, arc3, arc4

    def get_simple_model2(self):
        """

        Returns:

        """
        node1 = Storage(
            name="1",
            capacity=20,
            initial_storage={"volume": 10, "phosphate": 0.5, "temperature": 15},
        )
        node2 = Storage(
            name="2",
            capacity=30,
            initial_storage={"volume": 20, "phosphate": 0.2, "temperature": 12},
        )
        node3 = Node(name="3")

        node4 = Node(name="4")
        arc1 = Arc(in_port=node1, out_port=node3, name="arc1", preference=0.75)

        arc2 = Arc(in_port=node2, out_port=node3, name="arc2", preference=1.25)

        arc3 = Arc(in_port=node4, out_port=node1, name="arc3", preference=0.5)

        arc4 = Arc(in_port=node4, out_port=node2, name="arc4", preference=10)

        return node1, node2, node3, node4, arc1, arc2, arc3, arc4

    def test_inoutmb(self):
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}

        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}

        d3 = {
            "volume": 15,
            "phosphate": 0.00015,
            "temperature": (10 * 15 + 5 * 10) / 15,
        }

        d4 = {"volume": 0, "phosphate": 0, "temperature": 0}

        d5 = {"volume": 15, "phosphate": 0.00015, "temperature": 0}
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        arc2.send_push_request(d1)
        arc3.send_push_request(d2)

        in_, ds_, out_ = node1.node_mass_balance()

        self.assertDictAlmostEqual(d3, in_)
        self.assertDictAlmostEqual(d5, ds_)
        self.assertDictAlmostEqual(d4, out_)

        node1.distribute()

        in_ = node1.total_in()
        self.assertDictAlmostEqual(d3, in_)

        out_ = node1.total_out()
        self.assertDictAlmostEqual(d3, out_)

    def test_push(self):
        d1 = {"volume": 25, "phosphate": 0.0001, "temperature": 15}

        d2 = {"volume": 5, "phosphate": 0.00002, "temperature": 15}

        d3 = {"volume": 20, "phosphate": 0, "temperature": 0}

        d4 = {"volume": 10, "phosphate": 0.01, "temperature": 10}

        d5 = {"volume": 10, "phosphate": 0, "temperature": 0}
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        reply = node1.push_check()
        self.assertDictAlmostEqual(d3, reply)

        reply = node1.push_check(d4)
        self.assertDictAlmostEqual(d5, reply)

        reply = node1.push_set(d1)
        self.assertDictAlmostEqual(d2, reply)

    def test_pull(self):
        d1 = {"volume": 15, "phosphate": 0.0001, "temperature": 15}

        d2 = {"volume": 25}

        d3 = {"volume": 10}

        d4 = {"volume": 10, "phosphate": 0.0002 / 3, "temperature": 15}
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        node1.push_set(d1)

        reply = node1.pull_check()
        self.assertDictAlmostEqual(d1, reply)

        reply = node1.pull_check(d3)
        self.assertDictAlmostEqual(d4, reply)

        reply = node1.pull_set(d2)
        self.assertDictAlmostEqual(d1, reply)

    def test_direction(self):
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        f, outarcs = node1.get_direction_arcs("push")

        self.assertEqual("send_push_check", f)
        self.assertEqual(set([arc1, arc4]), set(outarcs))

        f, inarcs = node1.get_direction_arcs("pull")

        self.assertEqual("send_pull_check", f)
        self.assertEqual(set([arc3, arc2]), set(inarcs))

    def test_connected(self):
        d1 = {"volume": 15, "phosphate": 0.0001, "temperature": 15}

        connected_ = {
            "avail": 15,
            "priority": 7.5,
            "allocation": {"arc4": 7.5},
            "capacity": {"arc4": 15},
        }

        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        node1.push_set(d1)

        connected = node5.get_connected("pull")

        self.assertDictEqual(connected_, connected)

    def test_handler_push(self):
        d1 = {"volume": 25, "phosphate": 0.0001, "temperature": 15}

        d2 = {"volume": 5, "phosphate": 0.00002, "temperature": 15}

        d3 = {"volume": 20, "phosphate": 0, "temperature": 0}

        d4 = {"volume": 10, "phosphate": 0.01, "temperature": 10}

        d5 = {"volume": 10, "phosphate": 0, "temperature": 0}
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        reply = node1.query_handler(node1.push_check_handler, None, "default")
        self.assertDictAlmostEqual(d3, reply)

        reply = node1.query_handler(node1.push_check_handler, d4, "default")
        self.assertDictAlmostEqual(d5, reply)

        reply = node1.query_handler(node1.push_set_handler, d1, "default")
        self.assertDictAlmostEqual(d2, reply)

    def test_handler_pull(self):
        d1 = {"volume": 15, "phosphate": 0.0001, "temperature": 15}

        d2 = {"volume": 25}

        d3 = {"volume": 10}

        d4 = {"volume": 10, "phosphate": 0.0002 / 3, "temperature": 15}
        (
            node1,
            node2,
            node3,
            node4,
            node5,
            arc1,
            arc2,
            arc3,
            arc4,
        ) = self.get_simple_model1()

        node1.push_set(d1)

        reply = node1.query_handler(node1.pull_check_handler, None, "default")
        self.assertDictAlmostEqual(d1, reply)

        reply = node1.query_handler(node1.pull_check_handler, d3, "default")
        self.assertDictAlmostEqual(d4, reply)

        reply = node1.query_handler(node1.pull_set_handler, d2, "default")
        self.assertDictAlmostEqual(d1, reply)

    def test_pull_distributed(self):
        d1 = {"volume": 10}

        amount_from_1 = 10 * 10 * 0.75 / (10 * 0.75 + 20 * 1.25)
        amount_from_2 = 10 * 20 * 1.25 / (10 * 0.75 + 20 * 1.25)

        d2 = {
            "volume": 10,
            "phosphate": 0.5 * amount_from_1 / 10 + 0.2 * amount_from_2 / 20,
            "temperature": (15 * amount_from_1 + 12 * amount_from_2) / 10,
        }

        d3 = {
            "volume": amount_from_1,
            "phosphate": 0.5 * amount_from_1 / 10,
            "temperature": 15,
        }

        d4 = {
            "volume": amount_from_2,
            "phosphate": 0.2 * amount_from_2 / 20,
            "temperature": 12,
        }

        node1, node2, node3, node4, arc1, arc2, arc3, arc4 = self.get_simple_model2()

        reply = node3.pull_distributed(d1)
        self.assertDictAlmostEqual(d2, reply)
        self.assertDictAlmostEqual(d3, arc1.vqip_out)
        self.assertDictAlmostEqual(d4, arc2.vqip_out)

    def test_push_distributed(self):
        d1 = {"volume": 15, "phosphate": 0.01, "temperature": 12}

        d2 = {"volume": 0, "phosphate": 0, "temperature": 12}

        d3 = {"volume": 10, "phosphate": 0.01 * 2 / 3, "temperature": 12}

        d4 = {"volume": 5, "phosphate": 0.01 / 3, "temperature": 12}

        node1, node2, node3, node4, arc1, arc2, arc3, arc4 = self.get_simple_model2()
        reply = node4.push_distributed(d1)
        self.assertDictAlmostEqual(d2, reply)

        # Distributing is iterative so these won't be perfect
        self.assertDictAlmostEqual(d3, arc4.vqip_in, 14)
        self.assertDictAlmostEqual(d4, arc3.vqip_in, 14)

    def test_pull_check_basic(self):
        d1 = {"volume": 30, "phosphate": 0.7, "temperature": (15 * 10 + 12 * 20) / 30}

        d2 = {"volume": 10}

        d3 = {
            "volume": 10,
            "phosphate": 0.7 / 3,
            "temperature": (15 * 10 + 12 * 20) / 30,
        }

        node1, node2, node3, node4, arc1, arc2, arc3, arc4 = self.get_simple_model2()

        reply = node3.pull_check_basic()
        self.assertDictAlmostEqual(d1, reply)

        reply = node3.pull_check_basic(d2)
        self.assertDictAlmostEqual(d3, reply)

    def test_push_check_basic(self):
        d1 = {"volume": 20, "phosphate": 0.6, "temperature": 13.5}

        d2 = {"volume": 10, "phosphate": 0.1, "temperature": 17}

        d3 = {"volume": 10, "phosphate": 0.3, "temperature": 13.5}

        node1, node2, node3, node4, arc1, arc2, arc3, arc4 = self.get_simple_model2()
        reply = node4.push_check_basic()

        self.assertDictAlmostEqual(d1, reply)

        reply = node4.push_check_basic(d2)

        self.assertDictAlmostEqual(d3, reply)

    def test_deny(self):
        d1 = {"volume": 20, "phosphate": 0.6, "temperature": 13.5}

        d2 = {"volume": 0, "phosphate": 0, "temperature": 0}

        node = Node(name="")

        reply = node.pull_check_deny(d1)
        self.assertDictEqual(d2, reply)

        reply = node.pull_check_deny()
        self.assertDictEqual(d2, reply)

        reply = node.pull_set_deny(d1)
        self.assertDictEqual(d2, reply)

        reply = node.push_set_deny(d1)
        self.assertDictEqual(d1, reply)

        reply = node.push_check_deny(d1)
        self.assertDictEqual(d2, reply)

        reply = node.push_check_deny()
        self.assertDictEqual(d2, reply)

    def test_data_read(self):
        node = Node(name="", data_input_dict={("temperature", 1): 15})
        node.t = 1

        self.assertEqual(15, node.get_data_input("temperature"))

    def test_data_overrides(self):
        data_path = os.path.join(
            os.getcwd(),
            "docs",
            "demo",
            "data",
            "processed",
            "example_override_data.csv.gz",
        )
        input_data = pd.read_csv(data_path)

        overrides = {"data_input_dict": data_path}
        node = Node(name="")
        node.apply_overrides(overrides)
        node.t = list(node.data_input_dict.keys())[0][1]

        self.assertEqual(
            input_data.groupby("variable").get_group("temperature")["value"].iloc[0],
            node.get_data_input("temperature"),
        )
        # test runtime error
        self.assertRaises(RuntimeError, lambda: node.apply_overrides({}))


if __name__ == "__main__":
    unittest.main()



"""This module contains the utilities to extend WSMOD with new features.

The `register_node_patch` decorator is used to register a function that will be used
instead of a method or attribute of a node. The `apply_patches` function applies all
registered patches to a model.

Example of patching a method:

`empty_distributed` will be called instead of `pull_distributed` of "my_node":

    >>> from wsimod.extensions import register_node_patch, apply_patches
    >>> @register_node_patch("my_node", "pull_distributed")
    >>> def empty_distributed(self, vqip):
    >>>      return {}

Attributes, methods of the node, and sub-attributes can be patched. Also, an item of a
list or a dictionary can be patched if the item argument is provided.

Example of patching an attribute:

`10` will be assigned to `t`:

    >>> @register_node_patch("my_node", "t", is_attr=True)
    >>> def patch_t(node):
    >>>     return 10

Example of patching an attribute item:

`patch_default_pull_set_handler` will be assigned to
`pull_set_handler["default"]`:

    >>> @register_node_patch("my_node", "pull_set_handler", item="default")
    >>> def patch_default_pull_set_handler(self, vqip):
    >>>     return {}

If patching a method of an attribute, the `is_attr` argument should be set to `True` and
the target should include the attribute name and the method name, all separated by 
periods, eg. `attribute_name.method_name`.

It should be noted that the patched function should have the same signature as the
original method or attribute, and the return type should be the same as well, otherwise
there will be a runtime error. In particular, the first argument of the patched function
should be the node object itself, which will typically be named `self`.

The overridden method or attribute can be accessed within the patched function using the
`_patched_{method_name}` attribute of the object, eg. `self._patched_pull_distributed`. 
The exception to this is when patching an item, in which case the original item is no
available to be used within the overriding function.

Finally, the `apply_patches` is called within the `Model.load` method and will apply all
patches in the order they were registered. This means that users need to be careful with
the order of the patches in their extensions files, as they may have interdependencies.

TODO: Update documentation on extensions files.
"""

import warnings
from typing import Callable, Hashable

from .orchestration.model import Model

extensions_registry: dict[tuple[str, Hashable, bool], Callable] = {}


def register_node_patch(
    node_name: str, target: str, item: Hashable = None, is_attr: bool = False
) -> Callable:
    """Register a function to patch a node method or any of its attributes.

    Args:
        node_name (str): The name of the node to patch.
        target (str): The target of the object to patch in the form of a string with the
            attribute, sub-attribute, etc. and finally method (or attribute) to replace,
            sepparated with period, eg. `make_discharge` or
            `sewer_tank.pull_storage_exact`.
        item (Hashable): Typically a string or an integer indicating the item to replace
            in the selected attribue, which should be a list or a dictionary.
        is_attr (bool): If True, the decorated function will be called when applying
            the patch and the result assigned to the target, instead of assigning the
            function itself. In this case, the only argument passed to the function is
            the node object.
    """
    target_id = (node_name, target, item, is_attr)
    if target_id in extensions_registry:
        warnings.warn(f"Patch for {target} already registered.")

    def decorator(func):
        extensions_registry[target_id] = func
        return func

    return decorator


def apply_patches(model: Model) -> None:
    """Apply all registered patches to the model.

    TODO: Validate signature of the patched methods and type of patched attributes.

    Args:
        model (Model): The model to apply the patches to.
    """
    for (node_name, target, item, is_attr), func in extensions_registry.items():
        starget = target.split(".")
        method = starget.pop()

        # Get the member to patch
        node = obj = model.nodes[node_name]
        for attr in starget:
            obj = getattr(obj, attr)

        # Apply the patch
        if item is not None:
            obj = getattr(obj, method)
            obj[item] = func(node) if is_attr else func.__get__(node, node.__class__)
        else:
            setattr(obj, f"_patched_{method}", getattr(obj, method))
            setattr(
                obj, method, func(node) if is_attr else func.__get__(obj, obj.__class__)
            )



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
import warnings
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.sewer import Sewer
from wsimod.nodes.storage import (
    Groundwater,
    QueueGroundwater,
    Reservoir,
    River,
    RiverReservoir,
    Storage,
)
from wsimod.nodes.waste import Waste


class MyTestClass(TestCase):
    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_push_storage(self):
        constants.set_simple_pollutants()

        storage = Storage(
            name="",
            capacity=10,
            initial_storage={"volume": 4, "phosphate": 0.2, "temperature": 12},
        )

        d1 = {"volume": 6, "phosphate": 0.2 * 6 / 4, "temperature": 12}
        reply = storage.push_check()
        self.assertDictAlmostEqual(d1, reply)

        d2 = {"volume": 2, "phosphate": 0.2 * 2 / 4, "temperature": 12}
        reply = storage.push_check({"volume": 2})
        self.assertDictAlmostEqual(d2, reply)

        d3 = {"volume": 8, "phosphate": 0.25, "temperature": 13}
        reply = storage.push_set(d3)

        d4 = {"volume": 2, "phosphate": 0.25 * 2 / 8, "temperature": 13}
        self.assertDictAlmostEqual(d4, reply)

        d5 = {
            "volume": 10,
            "phosphate": 0.25 * 6 / 8 + 0.2,
            "temperature": (13 * 6 + 12 * 4) / 10,
        }
        self.assertDictAlmostEqual(d5, storage.tank.storage)

    def test_distribute(self):
        constants.set_simple_pollutants()
        d1 = {"volume": 4, "phosphate": 0.2, "temperature": 12}
        storage = Storage(name="", capacity=10, initial_storage=d1)
        waste = Waste(name="")
        arc1 = Arc(in_port=storage, out_port=waste, name="")

        storage.distribute()
        self.assertDictAlmostEqual(d1, arc1.vqip_in)
        d2 = {"volume": 0, "phosphate": 0, "temperature": 12}
        self.assertDictAlmostEqual(d2, storage.tank.storage)

    def test_groundwater_distribute(self):
        constants.set_simple_pollutants()
        d1 = {"volume": 4, "phosphate": 0.2, "temperature": 12}
        groundwater = Groundwater(
            name="", capacity=10, initial_storage=d1, residence_time=3
        )
        waste = Waste(name="")
        arc1 = Arc(in_port=groundwater, out_port=waste, name="")

        groundwater.distribute()
        d2 = {"volume": 4 / 3, "phosphate": 0.2 / 3, "temperature": 12}
        self.assertDictAlmostEqual(d2, arc1.vqip_in, 15)
        d2 = {"volume": 4 * 2 / 3, "phosphate": 0.2 * 2 / 3, "temperature": 12}
        self.assertDictAlmostEqual(d2, groundwater.tank.storage, 14)

    def test_groundwater_infiltrate(self):
        constants.set_simple_pollutants()
        d1 = {"volume": 4, "phosphate": 0.2, "temperature": 12}
        groundwater = Groundwater(
            name="",
            capacity=10,
            initial_storage=d1,
            infiltration_threshold=0.2,
            infiltration_pct=0.1,
        )
        sewer = Sewer(name="", capacity=5)
        arc1 = Arc(in_port=groundwater, out_port=sewer, name="")

        groundwater.infiltrate()
        nv = ((4 - (10 * 0.2)) * 0.1) ** 0.5
        d2 = {"volume": nv, "temperature": 12, "phosphate": 0.2 * nv / 4}
        self.assertDictAlmostEqual(d2, arc1.vqip_in)

    def test_qgroundwater_push(self):
        constants.set_simple_pollutants()
        groundwater = QueueGroundwater(name="", capacity=10, timearea={0: 0.2, 1: 0.8})
        d1 = {"volume": 4, "phosphate": 0.2, "temperature": 12}
        _ = groundwater.push_set_timearea(d1)
        self.assertDictAlmostEqual(d1, groundwater.tank.storage, 14)

        d2 = {"volume": 4 * 0.2, "phosphate": 0.2 * 0.2, "temperature": 12}
        self.assertDictAlmostEqual(d2, groundwater.tank.active_storage, 14)

        d3 = {"volume": 4 * 0.8, "phosphate": 0.2 * 0.8, "temperature": 12}
        self.assertDictAlmostEqual(d3, groundwater.tank.internal_arc.queue[1], 14)

        groundwater.end_timestep()
        self.assertDictAlmostEqual(d1, groundwater.tank.active_storage, 14)

    def test_qgroundwater_push(self):
        constants.set_simple_pollutants()
        groundwater = QueueGroundwater(name="", capacity=10, timearea={0: 0.2, 1: 0.8})

        waste = Waste(name="")
        arc1 = Arc(in_port=groundwater, out_port=waste, name="")

        d1 = {"volume": 4, "phosphate": 0.2, "temperature": 12}
        _ = groundwater.push_set_timearea(d1)

        groundwater.distribute()

        d2 = {"volume": 4 * 0.2, "phosphate": 0.2 * 0.2, "temperature": 12}
        self.assertDictAlmostEqual(d2, arc1.vqip_in, 14)

        d3 = {"volume": 4 * 0.8, "phosphate": 0.2 * 0.8, "temperature": 12}
        self.assertDictAlmostEqual(d3, groundwater.tank.storage, 14)

    def test_qgroundwater_pull(self):
        constants.set_simple_pollutants()
        groundwater = QueueGroundwater(name="", capacity=10, timearea={0: 0.2, 1: 0.8})

        waste = Waste(name="")
        Arc(in_port=groundwater, out_port=waste, name="")

        d1 = {"volume": 4, "phosphate": 0.2, "temperature": 12}
        _ = groundwater.push_set_timearea(d1)

        d2 = {"volume": 4 * 0.2, "phosphate": 0.2 * 0.2, "temperature": 12}
        self.assertDictAlmostEqual(d2, groundwater.pull_check_active(), 14)

        d2 = {"volume": 3, "phosphate": 0.2 * 3 / 4, "temperature": 12}
        reply = groundwater.pull_set_active({"volume": 3})
        self.assertDictAlmostEqual(d2, reply, 14)

        d3 = {
            "volume": 4 * 0.2 * (1 - 3 / 4),
            "phosphate": 0.2 * 0.2 * (1 - 3 / 4),
            "temperature": 12,
        }
        self.assertDictAlmostEqual(d3, groundwater.tank.active_storage, 14)

        d4 = {
            "volume": 4 * 0.8 * (1 - 3 / 4),
            "phosphate": 0.2 * 0.8 * (1 - 3 / 4),
            "temperature": 12,
        }
        self.assertDictAlmostEqual(d4, groundwater.tank.internal_arc.queue[1], 14)

    def test_river_pull(self):
        constants.set_simple_pollutants()
        river = River(
            name="",
            length=200,
            width=20,
            velocity=0.2 * 86400,
            damp=0.1,
            mrf=10,
            initial_storage={"volume": 30, "phosphate": 0.2, "temperature": 23},
        )
        d1 = {
            "volume": 19.988412514484356,
            "phosphate": 0.1332560834298957,
            "temperature": 23,
        }
        self.assertDictAlmostEqual(d1, river.pull_check_river())

        d2 = {"volume": 2.0, "phosphate": 0.2 * 2 / 30, "temperature": 23}
        self.assertDictAlmostEqual(d2, river.pull_check_river({"volume": 2}))

        reply = river.pull_set_river({"volume": 25})
        self.assertDictAlmostEqual(d1, reply)

    def test_river_depth(self):
        constants.set_simple_pollutants()
        river = River(
            name="",
            length=200,
            width=20,
            velocity=0.2 * 86400,
            damp=0.1,
            mrf=10,
            initial_storage={"volume": 30, "phosphate": 0.2, "temperature": 23},
        )
        river.update_depth()
        self.assertEqual(0.0075, river.current_depth)

    # TODO test river biochemical processes once they have been functionalised

    def test_reservoir(self):
        reservoir1 = Reservoir(
            name="",
            capacity=35,
            initial_storage={"volume": 30, "phosphate": 0.2, "temperature": 23},
        )
        d1 = {"volume": 3, "phosphate": 0.1, "temperature": 10}
        reservoir2 = Reservoir(name="", capacity=5, initial_storage=d1)

        arc1 = Arc(in_port=reservoir2, out_port=reservoir1, name="")

        reservoir1.make_abstractions()

        self.assertDictAlmostEqual(d1, arc1.vqip_in)

        d2 = {"volume": 33, "phosphate": 0.3, "temperature": (30 * 23 + 10 * 3) / 33}
        self.assertDictAlmostEqual(d2, reservoir1.tank.storage, 15)

    def test_riverreservoir_push(self):
        reservoir = RiverReservoir(
            name="",
            capacity=35,
            initial_storage={"volume": 30, "phosphate": 0.2, "temperature": 23},
            environmental_flow=5,
        )
        waste = Waste(name="")
        arc1 = Arc(in_port=reservoir, out_port=waste, arc="", capacity=20)

        reply = reservoir.push_check()
        d1 = {"volume": 25, "phosphate": 0.2 / 30 * 25, "temperature": 23}
        self.assertDictAlmostEqual(d1, reply, 15)

        reply = reservoir.push_set({"volume": 35, "phosphate": 0.5, "temperature": 20})
        d2 = {
            "volume": 10,
            "phosphate": (0.5 + 0.2) * 10 / 65,
            "temperature": (23 * 30 + 20 * 35) / 65,
        }
        self.assertDictAlmostEqual(d2, reply, 15)

        d3 = {
            "volume": 20,
            "phosphate": (0.5 + 0.2) * 20 / 65,
            "temperature": (23 * 30 + 20 * 35) / 65,
        }
        self.assertDictAlmostEqual(d3, arc1.vqip_in, 15)

    def test_riverreservoir_environmental(self):
        reservoir = RiverReservoir(
            name="",
            capacity=35,
            initial_storage={"volume": 30, "phosphate": 0.2, "temperature": 23},
            environmental_flow=15,
        )
        waste = Waste(name="")
        arc1 = Arc(in_port=reservoir, out_port=waste, arc="", capacity=20)

        reservoir.push_set({"volume": 10, "phosphate": 0.5, "temperature": 20})

        reservoir.satisfy_environmental()
        self.assertEqual(15, reservoir.total_environmental_satisfied)

        d1 = {
            "volume": 15,
            "phosphate": (0.5 + 0.2) * 15 / 40,
            "temperature": (20 * 10 + 30 * 23) / 40,
        }
        self.assertDictAlmostEqual(d1, arc1.vqip_in, 15)

        d2 = {
            "volume": 25,
            "phosphate": (0.5 + 0.2) * 25 / 40,
            "temperature": (20 * 10 + 30 * 23) / 40,
        }
        self.assertDictAlmostEqual(d2, reservoir.tank.storage, 15)

    def test_storage_overrides(self):
        constants.set_default_pollutants()
        storage = Storage(
            name="", decays={"phosphate": {"constant": 1.005, "exponent": 1.005}}
        )
        storage.apply_overrides(
            {
                "capacity": 1.43,
                "area": 2.36,
                "datum": 0.32,
                "decays": {"nitrite": {"constant": 10.005, "exponent": 1.105}},
            }
        )
        self.assertEqual(storage.capacity, 1.43)
        self.assertEqual(storage.tank.capacity, 1.43)
        self.assertEqual(storage.area, 2.36)
        self.assertEqual(storage.tank.area, 2.36)
        self.assertEqual(storage.datum, 0.32)
        self.assertEqual(storage.tank.datum, 0.32)
        self.assertDictEqual(
            storage.decays,
            {
                "phosphate": {"constant": 1.005, "exponent": 1.005},
                "nitrite": {"constant": 10.005, "exponent": 1.105},
            },
        )
        self.assertDictEqual(
            storage.tank.decays,
            {
                "phosphate": {"constant": 1.005, "exponent": 1.005},
                "nitrite": {"constant": 10.005, "exponent": 1.105},
            },
        )

    def test_groundwater_overrides(self):
        groundwater = Groundwater(
            name="", decays={"phosphate": {"constant": 1.005, "exponent": 1.005}}
        )
        groundwater.apply_overrides(
            {
                "residence_time": 27.3,
                "infiltration_threshold": 200,
                "infiltration_pct": 0.667,
                "capacity": 1.43,
                "area": 2.36,
                "datum": 0.32,
                "decays": {"nitrite": {"constant": 10.005, "exponent": 1.105}},
            }
        )
        self.assertEqual(groundwater.residence_time, 27.3)
        self.assertEqual(groundwater.infiltration_threshold, 200)
        self.assertEqual(groundwater.infiltration_pct, 0.667)
        self.assertEqual(groundwater.capacity, 1.43)
        self.assertEqual(groundwater.tank.capacity, 1.43)
        self.assertEqual(groundwater.area, 2.36)
        self.assertEqual(groundwater.tank.area, 2.36)
        self.assertEqual(groundwater.datum, 0.32)
        self.assertEqual(groundwater.tank.datum, 0.32)
        self.assertDictEqual(
            groundwater.decays,
            {
                "phosphate": {"constant": 1.005, "exponent": 1.005},
                "nitrite": {"constant": 10.005, "exponent": 1.105},
            },
        )
        self.assertDictEqual(
            groundwater.tank.decays,
            {
                "phosphate": {"constant": 1.005, "exponent": 1.005},
                "nitrite": {"constant": 10.005, "exponent": 1.105},
            },
        )

    def test_river_overrides(self):
        river = River(name="")
        overrides = {
            "length": 27.3,
            "width": 200,
            "velocity": 0.667,
            "damp": 1.43,
            "mrf": 2.36,
            "uptake_PNratio": 0.32,
            "bulk_density": 5.32,
            "denpar_w": 0.432,
            "T_wdays": 23.5,
            "halfsatINwater": 3.269,
            "hsatTP": 2.431,
            "limpppar": 6.473,
            "prodNpar": 7.821,
            "prodPpar": 8.231,
            "muptNpar": 6.213,
            "muptPpar": 7.021,
            "max_temp_lag": 3.213,
            "max_phosphorus_lag": 78.321,
        }
        overrides_to_check = overrides.copy()
        river.apply_overrides(overrides)
        for k, v in overrides_to_check.items():
            if k == "area":
                v = 27.3 * 200
                self.assertEqual(river.tank.area, v)
            if k == "capacity":
                v = constants.UNBOUNDED_CAPACITY
                self.assertEqual(river.tank.capacity, v)
            self.assertEqual(getattr(river, k), v)
        # test runtimeerrors
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            river.apply_overrides({"area": 75.2})
            assert "specifying area is depreciated" in str(w[0])

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            river.apply_overrides({"capacity": 123})
            assert "specifying capacity is depreciated" in str(w[0])

    def test_riverreservoir_overrides(self):
        riverreservoir = RiverReservoir(name="")
        riverreservoir.apply_overrides({"environmental_flow": 154})
        self.assertEqual(riverreservoir.environmental_flow, 154)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Created on Wed Apr  7 08:43:32 2021.

@author: Barney

Converted to totals on Thur Apr 21 2022
"""

from typing import Any, Dict

from wsimod.core import constants
from wsimod.core.core import DecayObj, WSIObj

# from wsimod.nodes import nodes #Complains about circular imports.
# I don't think it should do..


class Arc(WSIObj):
    """"""

    def __init__(
        self,
        name="",
        capacity=constants.UNBOUNDED_CAPACITY,
        preference=1,
        in_port=None,
        out_port=None,
        **kwargs,
    ):
        """Arc objects are the way for information to be passed between nodes in WSIMOD.
        They have an in_port (where a message comes from) and an out_port (where a
        message goes to).

        Returns:
            name (str): Name of arc. Defaults to ''.
            capacity (float): Capacity of flow along an arc (vol/timestep).
                Defaults to constants.UNBOUNDED_CAPACITY.
            preference (float): Number used to prioritise or deprioritise use of an arc
                when flexibility exists
            in_port: A WSIMOD node object where the arc starts
            out_port: A WSIMOD node object where the arc ends
        """
        # Default essential parameters
        self.name = name
        self.in_port = in_port
        self.out_port = out_port
        self.capacity = capacity
        self.preference = preference

        # Update args
        WSIObj.__init__(self)
        self.__dict__.update(kwargs)

        # def all_subclasses(cls):
        #     return set(cls.__subclasses__()).union(
        #         [s for c in cls.__subclasses__() for s in all_subclasses(c)])
        # node_types = [x.__name__ for x in all_subclasses(nodes.Node)] + ['Node']

        # if self.name in node_types:
        #     print('Warning: arc name should not take a node class name')
        #     #TODO... not sure why... also currently commented for import issues..

        # Initialise states
        self.flow_in = 0
        self.flow_out = 0
        self.vqip_in = self.empty_vqip()
        self.vqip_out = self.empty_vqip()

        # Update ports
        self.in_port.out_arcs[self.name] = self
        self.out_port.in_arcs[self.name] = self

        out_type = self.out_port.__class__.__name__
        in_type = self.in_port.__class__.__name__

        if hasattr(self.in_port, "out_arcs_type"):
            self.in_port.out_arcs_type[out_type][self.name] = self

        if hasattr(self.out_port, "in_arcs_type"):
            self.out_port.in_arcs_type[in_type][self.name] = self

        # Mass balance checking
        self.mass_balance_in = [lambda: self.vqip_in]
        self.mass_balance_out = [lambda: self.vqip_out]
        self.mass_balance_ds = [lambda: self.empty_vqip()]

    def apply_overrides(self, overrides: Dict[str, Any] = {}) -> None:
        """Apply overrides to the node.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.capacity = overrides.pop("capacity", self.capacity)
        self.preference = overrides.pop("preference", self.preference)
        if len(overrides) > 0:
            print(f"No override behaviour defined for: {overrides.keys()}")

    def arc_mass_balance(self):
        """Checks mass balance for inflows/outflows/storage change in an arc.

        Returns:
            in_ (dict) Total vqip of vqip_in and other inputs in mass_balance_in
            ds_ (dict): Total vqip of change in arc in mass_balance_ds
            out_ (dict): Total vqip of vqip_out and other outputs in mass_balance_out

        Examples:
            arc_in, arc_out, arc_ds = my_arc.arc_mass_balance()
        """
        in_, ds_, out_ = self.mass_balance()
        return in_, ds_, out_

    def send_push_request(self, vqip, tag="default", force=False):
        """Function used to transmit a push request from one node (in_port) to another
        node (out_port).

        Args:
            vqip (dict): A dict VQIP of water to push
            tag (str, optional):  optional message to direct the out_port's query_
                handler which function to call. Defaults to 'default'.
            force (bool, optional): Argument used to cause function to ignore tank
                capacity of out_port, possibly resulting in pooling. Should not be used
                    unless
                out_port is a tank object. Defaults to False.

        Returns:
            (dict): A VQIP amount of water that was not successfully pushed
        """
        vqip = self.copy_vqip(vqip)

        # Apply pipe capacity
        if force:
            not_pushed = self.empty_vqip()
        else:
            excess_in = self.get_excess(direction="push", vqip=vqip, tag=tag)
            not_pushed = self.v_change_vqip(
                vqip, max(vqip["volume"] - excess_in["volume"], 0)
            )

        # Don't attempt to send volume that exceeds capacity
        vqip = self.extract_vqip(vqip, not_pushed)

        # Set push
        reply = self.out_port.push_set(vqip, tag)

        # Update total amount successfully sent
        vqip = self.extract_vqip(vqip, reply)

        # Combine non-sent water
        reply = self.sum_vqip(reply, not_pushed)

        # Update mass balance
        self.flow_in += vqip["volume"]
        self.flow_out = self.flow_in

        self.vqip_in = self.sum_vqip(self.vqip_in, vqip)
        self.vqip_out = self.vqip_in

        return reply

    def send_pull_request(self, vqip, tag="default"):
        """Function used to transmit a pull request from one node (in_port) to another
        node (out_port).

        Args:
            vqip (dict): A dict VQIP of water to pull (by default, only 'volume' key is
                used)
            tag (str, optional): optional message to direct the out_port's query_handler
                which
                function to call. Defaults to 'default'.

        Returns:
            (dict): A VQIP amount of water that was successfully pulled
        """
        volume = vqip["volume"]
        # Apply pipe capacity
        excess_in = self.get_excess(direction="pull", vqip=vqip, tag=tag)["volume"]
        not_pulled = max(volume - excess_in, 0)
        volume -= not_pulled

        if volume > 0:
            for pol in constants.ADDITIVE_POLLUTANTS:
                if pol in vqip.keys():
                    vqip[pol] *= volume / vqip["volume"]

        vqip["volume"] = volume

        # Make pull
        vqip = self.in_port.pull_set(vqip, tag)

        # Update mass balance
        self.flow_in += vqip["volume"]
        self.flow_out = self.flow_in

        self.vqip_in = self.sum_vqip(self.vqip_in, vqip)
        self.vqip_out = self.vqip_in

        return vqip

    def send_push_check(self, vqip=None, tag="default"):
        """Function used to transmit a push check from one node (in_port) to another
        node (out_port).

        Args:
            vqip (dict): A dict VQIP of water to push that can be specified. Defaults to
                None, which returns maximum capacity to push.
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.

        Returns:
            (dict): A VQIP amount of water that could be pushed
        """
        return self.get_excess(direction="push", vqip=vqip, tag=tag)

    def send_pull_check(self, vqip=None, tag="default"):
        """Function used to transmit a pull check from one node (in_port) to another
        node (out_port).

        Args:
            vqip (dict): A dict VQIP of water to pull that can be specified (by default,
                only the 'volume' key is used). Defaults to None, which returns all
                    available water to pull.
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.

        Returns:
            (dict): A VQIP amount of water that could be pulled
        """
        return self.get_excess(direction="pull", vqip=vqip, tag=tag)

    def get_excess(self, direction, vqip=None, tag="default"):
        """Calculate how much could be pull/pulled along the arc by combining both arc
        capacity and out_port check information.

        Args:
            direction (str): should be 'pull' or 'push'
            vqip (dict, optional): A VQIP amount to push/pull that can be
                specified. Defaults to None, which returns all available water to
                pull or maximum capacity to push (depending on 'direction').
            tag (str, optional): optional message to direct the out_port's query_handler
                which function to call. Defaults to 'default'.

        Returns:
            (dict): A VQIP amount of water that could be pulled/pushed
        """
        # Pipe capacity
        pipe_excess = self.capacity - self.flow_in

        # Node capacity
        if direction == "push":
            node_excess = self.out_port.push_check(vqip, tag)
        elif direction == "pull":
            node_excess = self.in_port.pull_check(vqip, tag)
        excess = min(pipe_excess, node_excess["volume"])

        # TODO sensible to min(vqip, excess) here? (though it should be applied by node)

        return self.v_change_vqip(node_excess, excess)

    def end_timestep(self):
        """End timestep in an arc, resetting flow/vqip in/out (which determine) the
        capacity for that timestep."""
        self.vqip_in = self.empty_vqip()
        self.vqip_out = self.empty_vqip()
        self.flow_in = 0
        self.flow_out = 0

    def reinit(self):
        """Reinitiatilise."""
        self.end_timestep()


class QueueArc(Arc):
    """"""

    def __init__(self, number_of_timesteps=0, **kwargs):
        """A queue arc that stores each push or pull individually in the queue. Enables
        implementation of travel time. A fixed number of timesteps can be specified as a
        parameter, and additional number of timesteps can be specified when the requests
        are made.

        The queue is a list of requests, where their travel time is decremented
        by 1 each timestep. Any requests with a travel time of 0 will be sent
        onwards if the 'update_queue' function is called.

        Args:
            number_of_timesteps (int, optional): Fixed number of timesteps that
                it takes to traverse the arc. Defaults to 0.
        """
        self.number_of_timesteps = number_of_timesteps
        self.queue = []
        super().__init__(**kwargs)

        self.queue_storage = self.empty_vqip()
        self.queue_storage_ = self.empty_vqip()

        self.mass_balance_ds.append(lambda: self.queue_arc_ds())

    def queue_arc_ds(self):
        """Calculate change in amount of water and other pollutants in the arc.

        Returns:
            (dict): A VQIP amount of change
        """
        self.queue_storage = self.queue_arc_sum()
        return self.extract_vqip(self.queue_storage, self.queue_storage_)

    def queue_arc_sum(self):
        """Sum the total water in the requests in the queue of the arc.

        Returns:
            (dict): A VQIP amount of water/pollutants in the arc
        """
        queue_storage = self.empty_vqip()
        for request in self.queue:
            queue_storage = self.sum_vqip(queue_storage, request["vqip"])
        return queue_storage

    def send_pull_request(self, vqip, tag="default", time=0):
        """Function used to transmit a pull request from one node (in_port) to another
        node (out_port). Any pulled water is immediately removed from the out_port and
        then takes the travel time to be received. This function has not been
        extensively tested.

        Args:
            vqip (_type_): A dict VQIP of water to pull (by default, only 'volume' key
                is used)
            tag (str, optional): optional message to direct the out_port's query_handler
                which function to call. Defaults to 'default'.
            time (int, optional): Travel time for request to spend in the arc (in
                addition to the arc's 'number_of_timesteps' parameter). Defaults to 0.

        Returns:
            (dict): A VQIP amount of water that was successfully pulled.
        """
        volume = vqip["volume"]
        # Apply pipe capacity
        excess_in = self.get_excess(direction="pull", vqip=vqip)["volume"]
        not_pulled = max(volume - excess_in, 0)
        volume -= not_pulled

        for pol in constants.ADDITIVE_POLLUTANTS:
            if pol in vqip.keys():
                vqip[pol] *= volume / vqip["volume"]

        vqip["volume"] = volume

        # Make pull
        vqip = self.in_port.pull_set(vqip)

        # Update to queue request
        request = {"time": time + self.number_of_timesteps, "vqip": vqip}

        # vqtip enters arc as a request
        self.enter_queue(request, direction="pull")

        # Update request queue and return pulls from queue
        reply = self.update_queue(direction="pull")
        return reply

    def send_push_request(self, vqip_, tag="default", force=False, time=0):
        """Function used to transmit a push request from one node (in_port) to another
        node (out_port).

        Args:
            vqip_ (dict): A dict VQIP of water to push.
            tag (str, optional): optional message to direct the out_port's query_handler
                which function to call. Defaults to 'default'.
            force (bool, optional): Ignore the capacity of the arc (note does not
                currently, pass the force argument to the out_port). Defaults to False.
            time (int, optional): Travel time for request to spend in the arc (in
                addition to the arc's 'number_of_timesteps' parameter). Defaults to 0.

        Returns:
            (dict): A VQIP amount of water that was not successfully pushed
        """
        vqip = self.copy_vqip(vqip_)

        if vqip["volume"] < constants.FLOAT_ACCURACY:
            return self.empty_vqip()

        # Apply pipe capacity
        if force:
            not_pushed = self.empty_vqip()
        else:
            excess_in = self.get_excess(direction="push", vqip=vqip, tag=tag)
            not_pushed = self.v_change_vqip(
                vqip, max(vqip["volume"] - excess_in["volume"], 0)
            )

        vqip = self.extract_vqip(vqip, not_pushed)

        # Update to queue request
        request = {"time": time + self.number_of_timesteps, "vqip": vqip}

        # vqtip enters arc as a request
        self.enter_queue(request, direction="push", tag=tag)

        # Update request queue
        backflow = self.update_queue(direction="push")
        not_pushed = self.sum_vqip(not_pushed, backflow)

        if backflow["volume"] > vqip_["volume"]:
            print("more backflow than vqip...")

        self.vqip_in = self.extract_vqip(self.vqip_in, backflow)

        return not_pushed

    def enter_arc(self, request, direction, tag):
        """Function used to cause format a request into the format expected by the
        enter_queue function.

        Args:
            request (dict): A dict with a VQIP under the 'vqip' key and the travel
                time under the 'time' key.
            direction (str): Direction of flow, can be 'push' or 'pull
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.

        Returns:
            (dict): The request dict with additional information entered for the queue.
        """
        request["average_flow"] = request["vqip"]["volume"] / (request["time"] + 1)
        request["direction"] = direction
        request["tag"] = tag

        self.flow_in += request["average_flow"]
        self.vqip_in = self.sum_vqip(self.vqip_in, request["vqip"])

        return request

    def enter_queue(self, request, direction=None, tag="default"):
        """Add a request into the arc's queue list.

        Args:
            request (dict): A dict with a VQIP under the 'vqip' key and the travel
                time under the 'time' key.
            direction (str): Direction of flow, can be 'push' or 'pull
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.
        """
        # Update inflows and format request
        request = self.enter_arc(request, direction, tag)

        # Enter queue
        self.queue.append(request)

    def update_queue(self, direction=None, backflow_enabled=True):
        """Iterate over all requests in the queue, removing them if they have no volume.

        If a request is a push and has 0 travel time remaining then
        the push will be triggered at the out_port, if the out_port responds that
        it cannot receive the push, then this water will be returned as backflow
        (if enabled).

        If a request is a pull and has 0 travel time remaining then it is simply summed
        with other 0 travel time pull_requests and returned (since the pull is made at
        the out_port when the send_pull_request is made).


        Args:
            direction (str, optional): Direction of flow, can be 'push' or 'pull.
                Defaults to None.
            backflow_enabled (bool, optional): Enable backflow, described above, if not
                enabled then the request will remain in the queue until all water has
                been received. Defaults to True.

        Returns:
            total_backflow (dict): In the case of a push direction, any backflow will be
                returned as a VQIP amount
            total_removed (dict): In the case of a pull direction, any pulled water will
                be returned as a VQIP amount
        """
        done_requests = []

        total_removed = self.empty_vqip()
        total_backflow = self.empty_vqip()
        # Iterate over requests
        for request in self.queue:
            if request["direction"] == direction:
                vqip = request["vqip"]

                if vqip["volume"] < constants.FLOAT_ACCURACY:
                    # Add to queue for removal
                    done_requests.append(request)
                elif request["time"] == 0:
                    if direction == "push":
                        # Attempt to push request
                        reply = self.out_port.push_set(vqip, request["tag"])
                        removed = vqip["volume"] - reply["volume"]

                    elif direction == "pull":
                        # Water has already been pulled, so assume all received
                        removed = vqip["volume"]

                    else:
                        print("No direction")

                    # Update outflows
                    self.flow_out += request["average_flow"] * removed / vqip["volume"]
                    vqip_ = self.v_change_vqip(vqip, removed)
                    total_removed = self.sum_vqip(total_removed, vqip_)

                    # Assume that any water that cannot arrive at destination this
                    # timestep is backflow
                    rejected = self.v_change_vqip(vqip, vqip["volume"] - removed)

                    if backflow_enabled | (
                        rejected["volume"] < constants.FLOAT_ACCURACY
                    ):
                        total_backflow = self.sum_vqip(rejected, total_backflow)
                        done_requests.append(request)
                    else:
                        request["vqip"] = rejected

        self.vqip_out = self.sum_vqip(self.vqip_out, total_removed)

        # Remove done requests
        for request in done_requests:
            self.queue.remove(request)

        # return total_removed
        if direction == "pull":
            return total_removed
        elif direction == "push":
            return total_backflow
        else:
            print("No direction")

    def end_timestep(self):
        """End timestep in an arc, resetting flow/vqip in/out (which determine) the
        capacity for that timestep.

        Update times of requests in the queue.
        """
        self.vqip_in = self.empty_vqip()
        self.vqip_out = self.empty_vqip()
        self.flow_in = 0
        self.flow_out = 0

        self.queue_storage_ = self.copy_vqip(self.queue_storage)
        self.queue_storage = self.empty_vqip()

        for request in self.queue:
            request["time"] = max(request["time"] - 1, 0)

        # TODO - update_queue here?

    def reinit(self):
        """"""
        self.end_timestep()
        self.queue = []


class AltQueueArc(QueueArc):
    """"""

    def __init__(self, **kwargs):
        """A simpler queue arc that has a queue that is a dict where each key is the
        travel time.

        Cannot be used if arc capacity is dynamic. Cannot be used for pulls.
        """
        self.queue_arc_sum = self.alt_queue_arc_sum

        super().__init__(**kwargs)
        self.queue = {0: self.empty_vqip(), 1: self.empty_vqip()}
        self.max_travel = 1

    def alt_queue_arc_sum(self):
        """Sum the total water in the queue of the arc.

        Returns:
            (dict): A VQIP amount of water/pollutants in the arc
        """
        queue_storage = self.empty_vqip()
        for request in self.queue.values():
            queue_storage = self.sum_vqip(queue_storage, request)
        return queue_storage

    def enter_queue(self, request, direction="push", tag="default"):
        """Add a request into the arc's queue.

        Args:
            request (dict): A dict with a VQIP under the 'vqip' key and the travel
                time under the 'time' key.
            direction (str): Direction of flow, can be 'push' only. Defaults to 'push'
            tag (str, optional): Optional message for out_port's query handler, can be
                'default' only. Defaults to 'default'.
        """
        # Update inflows and format request
        request = self.enter_arc(request, direction, tag)

        # Sum into queue
        if request["time"] in self.queue.keys():
            self.queue[request["time"]] = self.sum_vqip(
                self.queue[request["time"]], request["vqip"]
            )
        else:
            self.queue[request["time"]] = request["vqip"]
            self.max_travel = max(self.max_travel, request["time"])

    def update_queue(self, direction=None, backflow_enabled=True):
        """Trigger the push of water in the 0th key for the queue, if the out_port
        responds that it cannot receive the push, then this water will be returned as
        backflow (if enabled).

        Args:
            direction (str): Direction of flow, can be 'push' only. Defaults to 'push'
            backflow_enabled (bool, optional): Enable backflow, described above, if not
                enabled then the request will remain in the queue until all water has
                been received. Defaults to True.

        Returns:
            backflow (dict): In the case of a push direction, any backflow will be
                returned as a VQIP amount
        """
        # TODO - can this work for pulls??

        total_removed = self.copy_vqip(self.queue[0])

        # Push 0 travel time water
        backflow = self.out_port.push_set(total_removed)

        if not backflow_enabled:
            self.queue[0] = backflow
            backflow = self.empty_vqip()
        else:
            self.queue[0] = self.empty_vqip()

        total_removed = self.v_change_vqip(
            total_removed, total_removed["volume"] - backflow["volume"]
        )

        self.flow_out += total_removed["volume"]
        self.vqip_out = self.sum_vqip(self.vqip_out, total_removed)

        return backflow

    def end_timestep(self):
        """End timestep in an arc, resetting flow/vqip in/out (which determine) the
        capacity for that timestep.

        Update timings in the queue.
        """
        self.vqip_in = self.empty_vqip()
        self.vqip_out = self.empty_vqip()
        self.flow_in = 0
        self.flow_out = 0
        self.queue_storage_ = self.copy_vqip(self.queue_storage)
        self.queue_storage = self.empty_vqip()

        queue_ = self.queue.copy()
        keys = self.queue.keys()
        for i in range(self.max_travel):
            if (i + 1) in keys:
                self.queue[i] = queue_[i + 1]
                self.queue[i + 1] = self.empty_vqip()

        self.queue[0] = self.sum_vqip(queue_[0], queue_[1])

    def reinit(self):
        """"""
        self.end_timestep()
        self.queue = {0: self.empty_vqip(), 1: self.empty_vqip()}


class DecayArc(QueueArc, DecayObj):
    """"""

    def __init__(self, decays={}, **kwargs):
        """A QueueArc that applies decays from a DecayObj.

        Args:
            decays (dict, optional): A dict of dicts containing a key for each pollutant
                that decays and within that, a key for each parameter (a constant and
                exponent). Defaults to {}.
        """
        self.decays = decays

        QueueArc.__init__(self, **kwargs)
        DecayObj.__init__(self, decays)

        self.mass_balance_out.append(lambda: self.total_decayed)

    def enter_queue(self, request, direction=None, tag="default"):
        """Add a request into the arc's queue list. Apply the make_decay function (i.e.,
        the decay that occur's this timestep).

        Args:
            request (dict): A dict with a VQIP under the 'vqip' key and the travel
                time under the 'time' key.
            direction (str): Direction of flow, can be 'push' or 'pull
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.
        """
        # Update inflows and format
        request = self.enter_arc(request, direction, tag)

        # TODO - currently decay depends on temp at the in_port data object..
        # surely on vqip would be more sensible? (though this is true in many
        # places including WTW)

        # Decay on entry
        request["vqip"] = self.make_decay(request["vqip"])

        # Append to queue
        self.queue.append(request)

    def end_timestep(self):
        """End timestep in an arc, resetting flow/vqip in/out (which determine) the
        capacity for that timestep.

        Update times of requests in the queue. Apply the make_decay function (i.e., the
        decay that occurs in the following timestep).
        """
        self.vqip_in = self.empty_vqip()
        self.vqip_out = self.empty_vqip()
        self.total_decayed = self.empty_vqip()
        self.flow_in = 0
        self.flow_out = 0

        self.queue_storage_ = self.copy_vqip(self.queue_storage)
        self.queue_storage = self.empty_vqip()

        for request in self.queue:
            request["vqip"] = self.make_decay(request["vqip"])
            request["time"] = max(request["time"] - 1, 0)


class DecayArcAlt(AltQueueArc, DecayObj):
    """"""

    def __init__(self, decays={}, **kwargs):
        """An AltQueueArc that applies decays from a DecayObj.

        Args:
            decays (dict, optional): A dict of dicts containing a key for each pollutant
                that decays and within that, a key for each parameter (a constant and
                exponent). Defaults to {}.
        """
        self.decays = {}

        # super().__init__(**kwargs)
        AltQueueArc.__init__(self, **kwargs)
        DecayObj.__init__(self, decays)

        self.end_timestep = self._end_timestep

        self.mass_balance_out.append(lambda: self.total_decayed)

    def enter_queue(self, request, direction=None, tag="default"):
        """Add a request into the arc's queue. Apply the make_decay function (i.e., the
        decay that occur's this timestep).

        Args:
            request (dict): A dict with a VQIP under the 'vqip' key and the travel
                time under the 'time' key.
            direction (str): Direction of flow, can be 'push' only. Defaults to 'push'
            tag (str, optional): Optional message for out_port's query handler, can be
                'default' only. Defaults to 'default'.
        """
        # TODO- has no tags

        # Update inflows and format
        request = self.enter_arc(request, direction, tag)

        # Decay on entry
        request["vqip"] = self.make_decay(request["vqip"])

        # Sum into queue
        if request["time"] in self.queue.keys():
            self.queue[request["time"]] = self.sum_vqip(
                self.queue[request["time"]], request["vqip"]
            )
        else:
            self.queue[request["time"]] = request["vqip"]
            self.max_travel = max(self.max_travel, request["time"])

    def _end_timestep(self):
        """End timestep in an arc, resetting flow/vqip in/out (which determine) the
        capacity for that timestep.

        Update timings in the queue. Apply the make_decay function (i.e., the decay that
        occurs in the following timestep).
        """
        self.vqip_in = self.empty_vqip()
        self.vqip_out = self.empty_vqip()
        self.total_decayed = self.empty_vqip()
        self.flow_in = 0
        self.flow_out = 0

        self.queue_storage_ = self.copy_vqip(self.queue_storage)
        self.queue_storage = (
            self.empty_vqip()
        )  # TODO I don't think this (or any queue_storage=  empty) is necessary

        queue_ = self.queue.copy()
        keys = self.queue.keys()
        for i in range(self.max_travel):
            if (i + 1) in keys:
                self.queue[i] = self.make_decay(queue_[i + 1])
                self.queue[i + 1] = self.empty_vqip()

        self.queue[0] = self.sum_vqip(self.queue[0], self.make_decay(queue_[0]))


class PullArc(Arc):
    """"""

    def __init__(self, **kwargs):
        """Subclass of Arc where pushes return no availability to push.

        This creates an Arc where only pull requests/checks can be sent, similar to a
        river abstraction.
        """
        super().__init__(**kwargs)
        self.send_push_request = self.send_push_deny
        self.send_push_check = self.send_push_check_deny

    def send_push_deny(self, vqip, tag="default", force=False):
        """Function used to deny any push requests.

        Args:
            vqip (dict): A dict VQIP of water to push
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.
            force (bool, optional): Argument used to cause function to ignore tank
                capacity of out_port, possibly resulting in pooling. Should not be used
                unless out_port is a tank object. Defaults to False.

        Returns:
            (dict): A VQIP amount of water that was not successfully pushed
        """
        return vqip

    def send_push_check_deny(self, vqip=None, tag="default"):
        """Function used to deny any push checks.

        Args:
            vqip (dict): A dict VQIP of water to push that can be specified. Defaults to
                None, which returns maximum capacity to push.
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.

        Returns:
            (dict): An empty VQIP amount of water indicating no water can be pushed
        """
        return self.empty_vqip()


class PushArc(Arc):
    """"""

    def __init__(self, **kwargs):
        """Subclass of Arc where pushes return no availability to pull.

        This creates an Arc where only push requests/checks can be sent, similar to a
        CSO.
        """
        super().__init__(**kwargs)
        self.send_pull_request = self.send_pull_deny
        self.send_pull_check = self.send_pull_check_deny

    def send_pull_deny(self, vqip, tag="default", force=False):
        """Function used to deny any pull requests.

        Args:
            vqip (dict): A dict VQIP of water to pull
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.
            force (bool, optional): Argument used to cause function to ignore tank
                capacity of out_port, possibly resulting in pooling. Should not be used
                unless  out_port is a tank object. Defaults to False.

        Returns:
            (dict): A VQIP amount of water that was successfully pulled
        """
        return self.empty_vqip()

    def send_pull_check_deny(self, vqip=None, tag="default"):
        """Function used to deny any pull checks.

        Args:
            vqip (dict): A dict VQIP of water to pull that can be specified. Defaults to
                None, which returns maximum capacity to pull.
            tag (str, optional):  optional message to direct the out_port's
                query_handler which function to call. Defaults to 'default'.

        Returns:
            (dict): An empty VQIP amount of water indicating no water can be pulled
        """
        return self.empty_vqip()


class SewerArc(Arc):
    """"""

    pass


class WeirArc(SewerArc):
    """"""

    pass



# -*- coding: utf-8 -*-
"""Created on Sun Aug 14 16:27:14 2022.

@author: bdobson
"""

from typing import Any, Dict

from wsimod.core import constants
from wsimod.nodes.nodes import Node


def decorate_leakage_set(self, f):
    """Decorator to extend the functionality of `f` by introducing leakage. This is
    achieved by adjusting the volume of the request (vqip) to include anticipated
    leakage, calling the original function `f`, and then distributing the leaked amount
    to groundwater.

    Args:
        self (instance of Distribution class): The Distribution object to be
            extended
        f (function): The function to be extended. Expected to be the
            Distribution object's pull_set function.

    Returns:
        pull_set (function): The decorated function which includes the
            original functionality of `f` and additional leakage operations.
    """

    def pull_set(vqip, **kwargs):
        """

        Args:
            vqip:
            **kwargs:

        Returns:

        """
        vqip["volume"] /= 1 - self.leakage

        reply = f(vqip, **kwargs)

        amount_leaked = self.v_change_vqip(reply, reply["volume"] * self.leakage)

        reply = self.extract_vqip(reply, amount_leaked)

        unsuccessful_leakage = self.push_distributed(
            amount_leaked, of_type="Groundwater"
        )
        if unsuccessful_leakage["volume"] > constants.FLOAT_ACCURACY:
            print(
                "warning, distribution leakage not going to GW in {0} at {1}".format(
                    self.name, self.t
                )
            )
            reply = self.sum_vqip(reply, unsuccessful_leakage)

        return reply

    return pull_set


def decorate_leakage_check(self, f):
    """Decorator to extend the functionality of `f` by introducing leakage. This is
    achieved by adjusting the volume of the request (vqip) to include anticipated
    leakage and then calling the original function `f`.

    Args:
        self (instance of Distribution class): The Distribution object to be
            extended
        f (function): The function to be extended. Expected to be the
            Distribution object's pull_set function.

    Returns:
        pull_check (function): The decorated function which includes the
            original functionality of `f` and additional leakage operations.
    """

    def pull_check(vqip, **kwargs):
        """

        Args:
            vqip:
            **kwargs:

        Returns:

        """
        if vqip is not None:
            vqip["volume"] /= 1 - self.leakage
        reply = f(vqip, **kwargs)
        amount_leaked = self.v_change_vqip(reply, reply["volume"] * self.leakage)

        reply = self.extract_vqip(reply, amount_leaked)
        return reply

    return pull_check


class Distribution(Node):
    """"""

    def __init__(self, leakage=0, **kwargs):
        """A Node that cannot be pushed to. Intended to pass calls to FWTW - though this
        currently relies on the user to connect it properly.

        Args:
            leakage (float, optional): 1 > float >= 0 to express how much
                water should be leaked to any attached groundwater nodes. This
                number represents the proportion of total flow through the node
                that should be leaked.
                Defaults to 0.

        Functions intended to call in orchestration:
            None

        Key assumptions:
            - No distribution processes yet represented, this class is just
                for conveyance.

        Input data and parameter requirements:
            - None
        """
        self.leakage = leakage
        super().__init__(**kwargs)
        # Update handlers
        self.push_set_handler["default"] = self.push_set_deny
        self.push_check_handler["default"] = self.push_check_deny
        self.decorate_pull_handlers()

    def decorate_pull_handlers(self):
        """Decorate handlers if there is leakage ratio."""
        if self.leakage > 0:
            self.pull_set_handler["default"] = decorate_leakage_set(
                self, self.pull_set_handler["default"]
            )
            self.pull_check_handler["default"] = decorate_leakage_check(
                self, self.pull_check_handler["default"]
            )

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the sewer.

        Enables a user to override any of the following parameters:
        leakage.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.leakage = overrides.pop("leakage", self.leakage)
        self.decorate_pull_handlers()
        super().apply_overrides(overrides)


class UnlimitedDistribution(Distribution):
    """"""

    def __init__(self, **kwargs):
        """A distribution node that provides unlimited water while tracking pass
        balance.

        Functions intended to call in orchestration:
            None

        Key assumptions:
            - Water demand is always satisfied.

        Input data and parameter requirements:
            - None
        """
        super().__init__(**kwargs)
        # Update handlers
        self.pull_set_handler["default"] = self.pull_set_unlimited
        self.pull_check_handler["default"] = lambda x: self.v_change_vqip(
            self.empty_vqip(), constants.UNBOUNDED_CAPACITY
        )

        # States
        self.supplied = self.empty_vqip()

        self.mass_balance_in.append(lambda: self.supplied)

    def pull_set_unlimited(self, vqip):
        """Respond that VQIP was fulfilled and update state variables for mass balance.

        Args:
            vqip (dict): A VQIP amount to request

        Returns:
            vqip (dict): A VQIP amount that was supplied
        """
        # TODO maybe need some pollutant concentrations?
        vqip = self.v_change_vqip(self.empty_vqip(), vqip["volume"])
        self.supplied = self.sum_vqip(self.supplied, vqip)
        return vqip

    def end_timestep(self):
        """Update state variables."""
        self.supplied = self.empty_vqip()



# -*- coding: utf-8 -*-
"""Created on Fri Dec  6 15:17:07 2019.

@author: bdobson
"""
from wsimod.core import constants

M3_S_TO_ML_D = 86.4
MM_KM2_TO_ML = 1e-3 * 1e6 * 1e3 * 1e-6  # mm->m, km2->m2, m3->l, l->Ml
MM_M2_TO_ML = 1e-3 * 1e3 * 1e-6  # mm->m, m3->l, l->Ml
MM_M2_TO_SIM_VOLUME = MM_M2_TO_ML  # SIM volume is by default ML, but can be changed by
# changing MM_TO_SIM_VOLUME
MM_M2_TO_M3 = 1e-3  # mm->m
ML_TO_M3 = 1000
PCT_TO_PROP = 1 / 100
L_TO_ML = 1e-6
FLOAT_ACCURACY = 1e-11
UNBOUNDED_CAPACITY = 1e15
MAX_INFLOW = 1e10
DT_DAYS = 1
POLLUTANTS = [
    "do",
    "org-phosphorus",
    "phosphate",
    "ammonia",
    "solids",
    "bod",
    "cod",
    "ph",
    "temperature",
    "nitrate",
    "nitrite",
    "org-nitrogen",
]
NON_ADDITIVE_POLLUTANTS = [
    "do",
    "temperature",
    "ph",
]  # e.g. pollutants whose concentration should not increase if volume is distilled out
ADDITIVE_POLLUTANTS = [
    "org-phosphorus",
    "phosphate",
    "ammonia",
    "solids",
    "bod",
    "cod",
    "nitrate",
    "nitrite",
    "org-nitrogen",
]
NUTRIENTS = ["N", "P"]
PCT_GARDENS = 0.1  # Percentage of area that is people's gardens
PI = 3.141592653589793
PER_DAY_TO_PER_SECOND = 1 / (60 * 60 * 24)
PER_DAY_TO_PER_HOUR = 1 / 24
M3_S_TO_ML_H = M3_S_TO_ML_D * PER_DAY_TO_PER_HOUR
MAXITER = 5  # Max iterations in a while loop
M3_S_TO_M3_DT = 86400
M_S_TO_M_DT = 86400
UG_L_TO_KG_M3 = 1e-6
MG_L_TO_KG_M3 = 1e-3
KG_M3_TO_MG_L = 1e3
MM_TO_M = 1e-3
G_TO_KG = 1e-3
KG_TO_MG = 1e6
DECAY_REFERENCE_TEMPERATURE = 20  # C
PER_30MONTH_TO_PER_DAY = 1 / 30
DAYS_IN_YEAR = 365
L_S_TO_M3_D = 1e-3 / PER_DAY_TO_PER_SECOND

M_TO_MM = 1e3
MM_TO_M = 1e-3
KM_TO_M = 1e3
M_TO_KM = 1e-3
MJ_M2_TO_CAL_CM2 = 23.889
KPA_TO_MBAR = 10
HPA_TO_KPA = 0.1
KM2_TO_HA = 100
KM2_TO_M2 = 1e6
HA_TO_M2 = 1e4
M2_TO_KM2 = 1e-6

G_M2_TO_KG_KM2 = 1e3
G_M2_TO_KG_M2 = 1e-3
MGMM_L_TO_KG_KM2 = 1e-6 * 1e3 * 1e-3 * 1e6
KG_M2_TO_KG_KM2 = 1e6

KG_M3_TO_KG_KM3 = 1e9

D_TO_S = 3600 * 24


def set_simple_pollutants():
    """"""
    constants.POLLUTANTS = ["phosphate", "temperature"]
    constants.ADDITIVE_POLLUTANTS = ["phosphate"]
    constants.NON_ADDITIVE_POLLUTANTS = ["temperature"]


def set_default_pollutants():
    """"""
    constants.POLLUTANTS = [
        "do",
        "org-phosphorus",
        "phosphate",
        "ammonia",
        "solids",
        "bod",
        "cod",
        "ph",
        "temperature",
        "nitrate",
        "nitrite",
        "org-nitrogen",
    ]
    constants.NON_ADDITIVE_POLLUTANTS = [
        "do",
        "temperature",
        "ph",
    ]  # e.g. pollutants whose concentration should not increase if volume is distilled
    # out
    constants.ADDITIVE_POLLUTANTS = [
        "org-phosphorus",
        "phosphate",
        "ammonia",
        "solids",
        "bod",
        "cod",
        "nitrate",
        "nitrite",
        "org-nitrogen",
    ]



# -*- coding: utf-8 -*-
"""Created on Mon Nov 15 14:20:36 2021.

@author: bdobson
"""
from wsimod.nodes.nodes import Node


# TODO call this outlet not waste
class Waste(Node):
    """"""

    def __init__(self, name):
        """Outlet node that can receive any amount of water by pushes.

        Args:
            name (str): Node name

        Functions intended to call in orchestration:
            None

        Key assumptions:
            - Water 'disappears' (leaves the model) from these nodes.

        Input data and parameter requirements:
            - None
        """
        # Update args
        super().__init__(name)

        # Update handlers
        self.pull_set_handler["default"] = self.pull_set_deny
        self.pull_check_handler["default"] = self.pull_check_deny
        self.push_set_handler["default"] = self.push_set_accept
        self.push_check_handler["default"] = self.push_check_accept

        # Mass balance
        self.mass_balance_out.append(self.total_in)

    def push_set_accept(self, vqip):
        """Push set function that accepts all water.

        Args:
            vqip (dict): A VQIP that has been pushed (ignored)

        Returns:
            (dict): An empty VQIP, indicating all water was received
        """
        return self.empty_vqip()



# -*- coding: utf-8 -*-
"""Created on Mon Nov 15 14:20:36 2021.

@author: bdobson
Converted to totals on 2022-05-03
"""
from typing import Any, Dict

from wsimod.core import constants
from wsimod.nodes.nodes import Node
from wsimod.nodes.tanks import QueueTank


class Sewer(Node):
    """"""

    def __init__(
        self,
        name,
        capacity=0,
        pipe_time=0,  # Sewer to sewer travel time
        pipe_timearea={0: 1},
        chamber_area=1,
        chamber_floor=10,
        data_input_dict={},
    ):
        """Sewer node that has a QueueTank and storage capacity. Think carefully about
        parameterising this tank, because of course the amount of water that can flow
        through a sewer in a timestep is different in reality than in a.

        steady state (e.g., a sewer that can handle a peak of 6m3/s in practice could
        not handle 6 * 86400 m3 of water in a day because that water does not flow
        uniformly over the day).

        Args:
            name (str): node name
            capacity (float, optional): Sewer tank capacity. Defaults to 0.
            pipe_time (float, optional): Number of timesteps to spend in the queue of
                the sewer tank. Defaults to 0.
            pipe_timearea (dict, optional): Time area diagram that enables flows to
                take a range of different durations to 'traverse' the tank. The keys
                of the dict are the number of timesteps while the values are the
                proportion of flow. E.g., {0 : 0.7, 1 : 0.3} means 70% of flow takes
                0 timesteps and 30% takes 1 timesteps.
            chamber_area (float, optional): Sewer tank area. Defaults to 1.
            chamber_floor (float, optional): Sewer tank datum. Defaults to 10.
            data_input_dict (dict, optional): Dictionary of data inputs relevant for
                the node (though I don't think it is used). Defaults to {}.

        NOTE that currently the queuetank either applies the pipe_timearea
        (push_set_land) OR the pipe_time (push_set_sewer). Though this behaviour
        could be changed by setting the number_of_timesteps property to pipe_time of
        the sewer_tank and removing the pipe_time setting in push_set_sewer.

        Functions intended to call in orchestration:
            make_discharge

        Key assumptions:
            - Sewer networks can be represented in an aggregated manner, where
                the behaviour of collections of manholes/pipes can be captured
                in a single component.
            - Travel time of water received from either `land.py/Land` objects
                or `demand.py/Demand` objects is assumed to be received as a
                non-point source and thus can be represented with the time-area
                method.
            - Travel time of water from an upstream `Sewer` object has a fixed
                travel time through the node.
            - The flow capacity of sewer network can be represented as with a
                `Tank`.
            - The `Sewer` object is not currently biochemically active.

        Input data and parameter requirements:
            - `pipe_timearea` is a dictionary containing the timearea diagram.
                _Units_: duration of flow (in timesteps) and proportion of flow
            - `pipe_time` describes the travel time of water received from upstream
                `Sewer`
                objects.
                _Units_: number of timesteps
            - `capacity`, `chamber_area`, `chamber_datum` describe the dimensions of the
                `Tank` that controls flow.
                _Units_: cubic metres, squared metres, metres
        """
        # Set parameters
        self.capacity = capacity
        self.pipe_time = pipe_time
        self.pipe_timearea = pipe_timearea
        self.chamber_area = chamber_area
        self.chamber_floor = chamber_floor
        # TODO I don't think this is used..
        self.data_input_dict = data_input_dict

        # Update args
        super().__init__(name)

        # Update handlers
        self.push_set_handler["Sewer"] = self.push_set_sewer
        self.push_set_handler["default"] = self.push_set_sewer
        self.push_set_handler["Land"] = self.push_set_land
        self.push_set_handler["Demand"] = self.push_set_land

        self.push_check_handler["default"] = self.push_check_sewer
        self.push_check_handler["Sewer"] = self.push_check_sewer
        self.push_check_handler["Demand"] = self.push_check_sewer
        self.push_check_handler["Land"] = self.push_check_sewer

        # Create sewer tank
        # TODO this might work better as a ResidenceTank (maybe also decay?)
        self.sewer_tank = QueueTank(
            capacity=self.capacity,
            number_of_timesteps=0,
            datum=self.chamber_floor,
            area=self.chamber_area,
        )

        # Mass balance
        self.mass_balance_ds.append(lambda: self.sewer_tank.ds())

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the sewer.

        Enables a user to override any of the following parameters:
        capacity, chamber_area, chamber_floor, pipe_time, pipe_timearea.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.capacity = overrides.pop("capacity", self.capacity)
        self.chamber_area = overrides.pop("chamber_area", self.chamber_area)
        self.chamber_floor = overrides.pop("chamber_floor", self.chamber_floor)
        self.sewer_tank.capacity = self.capacity
        self.sewer_tank.area = self.chamber_area
        self.sewer_tank.datum = self.chamber_floor

        self.pipe_time = overrides.pop("pipe_time", self.pipe_time)
        if "pipe_timearea" in overrides.keys():
            pipe_timearea_sum = sum([v for k, v in overrides["pipe_timearea"].items()])
            if pipe_timearea_sum != 1:
                print(
                    "ERROR: the sum of pipe_timearea in the overrides dict \
			is not equal to 1, please check it"
                )
        self.pipe_timearea = overrides.pop("pipe_timearea", self.pipe_timearea)
        super().apply_overrides(overrides)

    def push_check_sewer(self, vqip=None):
        """Generic push check, simply looks at excess.

        Args:
            vqip (dict, optional): A VQIP that can be used to limit the volume in
                the return value (only volume key is used). Defaults to None.

        Returns:
            excess (dict): Sewer tank excess
        """
        # Get excess
        excess = self.sewer_tank.get_excess()
        if vqip is None:
            return excess
        # Limit respone to vqip volume
        excess = self.v_change_vqip(excess, min(excess["volume"], vqip["volume"]))
        return excess

    def push_set_sewer(self, vqip):
        """Generic push request setting that implements basic queue travel time (it does
        NOT implement timearea travel time). Updates the sewer tank storage. Assumes
        that the inflow arc has accurately calculated capacity with push_check_sewer,
        thus the water is forced.

        Args:
            vqip (dict): A VQIP amount of water to push

        Returns:
            (dict): A VQIP amount of water that was not received
        """
        # Sewer to sewer push, update queued tank
        return self.sewer_tank.push_storage(vqip, time=self.pipe_time)

    def push_set_land(self, vqip):
        """Push request that applies pipe_timearea (see __init__ for description). As
        with push_set_sewer, push is also forced. Used to receive flow from land or
        demand that is assumed to occur widely across some kind of sewer catchment.

        Args:
            vqip (dict): A VQIP amount to be pushed

        Returns:
            (dict): A VQIP amount that was not received
        """
        # Land/demand to sewer push, update queued tank

        reply = self.empty_vqip()

        # Iterate over timearea diagram
        for time, normalised in self.pipe_timearea.items():
            vqip_ = self.v_change_vqip(vqip, vqip["volume"] * normalised)
            reply_ = self.sewer_tank.push_storage(vqip_, time=time)
            reply = self.sum_vqip(reply, reply_)

        return reply

    def make_discharge(self):
        """Function to trigger downstream sewer flow.

        Updates sewer tank travel time, pushes to WWTW, then sewer, then CSO. May flood
        land if, after these attempts, the sewer tank storage is above capacity.
        """
        self.sewer_tank.internal_arc.update_queue(direction="push")
        # TODO... do I need to do anything with this backflow... does it ever happen?
        # Discharge to Sewer if possible
        # remaining = self.push_distributed(self.sewer_tank.active_storage,
        #                                 of_type = 'Sewer',
        #                                 tag = 'Sewer')

        # #Discharge to WWTW if possible
        # remaining = self.push_distributed(remaining,
        #                                 of_type = 'WWTW',
        #                                 tag = 'Sewer')

        # #CSO discharge
        # remaining = self.push_distributed(remaining,
        #                                   of_type = ['Node', 'River'])

        remaining = self.push_distributed(self.sewer_tank.active_storage)

        # TODO backflow can cause mass balance errors here

        # Update tank
        sent = self.extract_vqip(self.sewer_tank.active_storage, remaining)
        reply = self.sewer_tank.pull_storage_exact(sent)
        if (reply["volume"] - sent["volume"]) > constants.FLOAT_ACCURACY:
            print("Miscalculated tank storage in discharge")

        # Flood excess
        ponded = self.sewer_tank.pull_ponded()
        if ponded["volume"] > constants.FLOAT_ACCURACY:
            reply_ = self.push_distributed(ponded, of_type=["Land"], tag="Sewer")
            reply_ = self.sewer_tank.push_storage(reply_, time=0, force=True)
            if reply_["volume"]:
                print("ponded water cant reenter")

    def end_timestep(self):
        """Overwrite end_timestep behaviour to update tank variables."""
        self.sewer_tank.end_timestep()

    def reinit(self):
        """Call Tank reinit."""
        self.sewer_tank.reinit()


class EnfieldFoulSewer(Sewer):
    """"""

    # TODO: combine with sewer
    def __init__(
        self,
        name,
        capacity=0,
        pipe_time=0,  # Sewer to sewer travel time
        pipe_timearea={0: 1},
        chamber_area=1,
        chamber_floor=10,
        data_input_dict={},
    ):
        """Alternate legacy sewer class...

        I dont think this is needed any more.
        """
        # TODO above

        super().__init__(
            name,
            capacity=capacity,
            pipe_time=pipe_time,
            pipe_timearea=pipe_timearea,
            chamber_area=chamber_area,
            chamber_floor=chamber_floor,
            data_input_dict=data_input_dict,
        )
        self.__class__.__name__ = "Sewer"

    def make_discharge(self):
        """"""
        _ = self.sewer_tank.internal_arc.update_queue(direction="push")

        # Discharge downstream
        if (
            self.sewer_tank.storage["volume"]
            > self.storm_exchange * self.sewer_tank.capacity
        ):
            exchange_v = min(
                (1 - self.storm_exchange) * self.sewer_tank.capacity,
                self.sewer_tank.active_storage["volume"],
            )
            exchange = self.v_change_vqip(self.sewer_tank.active_storage, exchange_v)
            remaining = self.push_distributed(exchange)
            sent_to_exchange = self.v_change_vqip(
                self.sewer_tank.active_storage, exchange_v - remaining["volume"]
            )
            self.sewer_tank.pull_storage(sent_to_exchange)

        remaining = self.push_distributed(
            self.sewer_tank.active_storage, of_type=["Waste"]
        )

        # Update tank
        sent = self.sewer_tank.active_storage["volume"] - remaining["volume"]
        sent = self.v_change_vqip(self.sewer_tank.active_storage, sent)
        reply = self.sewer_tank.pull_storage(sent)
        if (reply["volume"] - sent["volume"]) > constants.FLOAT_ACCURACY:
            print("Miscalculated tank storage in discharge")



# -*- coding: utf-8 -*-
"""Created on Mon Oct 31 15:14:51 2022.

@author: Barney

"""
import os
import unittest

loader = unittest.TestLoader()
start_dir = os.path.abspath("")
suite = loader.discover(start_dir)

runner = unittest.TextTestRunner()
runner.run(suite)



from pathlib import Path
from typing import List

from pytest import mark


def available_tutorials() -> List[Path]:
    """Locate the available tutorials in the docs."""
    base_path = Path(__file__).parent.parent / "docs" / "demo"
    return [p for p in base_path.rglob("*.py")]


@mark.parametrize("tutorial", available_tutorials())
def test_tutorials(tutorial):
    import subprocess
    import sys

    subprocess.run([sys.executable, tutorial]).check_returncode()



# %% [markdown]
# # Customise interactions (.py)
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/demo/scripts](https://github.com/imperialcollegelondon/wsi/blob/main/docs/demo/scripts/customise_interactions.py)
#
# 1. [Introduction](#introduction)
#
# 2. [Illustration](#illustration)
#
# 3. [More realistic example](#more-realistic-example)
#
# 4. [What next](#what-next)
#
# %% [markdown]
# ## Introduction
# In this tutorial we will demonstrate how to customise interactions between
# nodes using the handler functionality. Custom handlers are to be used when
# you want for a node to respond differently to one type of node than another.
# An example of custom handlers can be found in the
# [sewer node](https://github.com/imperialcollegelondon/wsi/blob/main/wsimod/nodes/sewer.py)
# where you can see that the sewer uses different functions to respond to
# differently tagged push requests. We highlight the `push_set_handler` customisations
# below for a sewer object below.
#
# Following the [extensions release] the most efficient way to customise anything
# within WSIMOD is to store your customisations in separate modules which are
# registered and applied when the model is initialised. Thus, to incrementally
# build up the example we will store each customisation separately in its own module.
# However, in practice one would contain these in a single module.
#
# But first...
#
# ## What are handlers?
# %%
from pprint import pprint as print

from wsimod.nodes.sewer import Sewer


my_sewer = Sewer(name="mr_sewer")
print(my_sewer.push_set_handler)

# %% [markdown]
# We can see that `my_sewer` has two different functions (`push_set_land`
# and `push_set_sewer`) stored in its `push_set_handler` dictionary. There are
# four different entries in the handler (`Demand`, `Land`, `Sewer`,
# `default`). This means that the `Sewer` node can respond to these four
# different tags. Tags can take any value, but by convention they specify a
# WSIMOD node type, or a tuple containing a node type and a specific type of
# interaction (e.g., `('Demand','Garden')` in the
# [land node](https://imperialcollegelondon.github.io/wsi/reference-land/#wsimod.nodes.land.Land)).
# All nodes must have the dictionaries: `push_set_handler`, `push_check_handler`,
# `pull_set_handler`, `pull_check_handler`, and at least one `default` key in each
# (see the [Node](https://imperialcollegelondon.github.io/wsi/reference-nodes/#wsimod.nodes.nodes.Node)
# class for defaults). If you do not want to define behaviour for a certain kind of interaction,
# e.g., maybe you can never pull from this type of node, then you can use, e.g.,
# the `pull_set_deny` and `pull_check_deny` functions for the `default` tag.
#
# It is important to note that unless a tag is specified in the request, the `default`
# tag is used.
# %% [markdown]
# ## Illustration
# We can illustrate the behaviour of handlers by creating a simple model. We will
# create a temporary directory to save our model files in for the purpose of this
# tutorial.
# %%
# Import packages
import os
import tempfile

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes import Distribution, Node
from wsimod.orchestration.model import Model, to_datetime

# Identify the location of the scripts folder
scripts_folder = os.path.join(os.path.abspath(""), "docs", "demo", "scripts")

# Create temporary directory
temp_dir = tempfile.TemporaryDirectory()

# Set simple pollutants (i.e., temperature and phosphate only)
constants.set_simple_pollutants()

# Create objects
my_node = Node(name="my_node")
my_dist = Distribution(name="my_dist")
my_arc = Arc(name="my_arc", in_port=my_node, out_port=my_dist)

# Wrap in a model for convenience
my_model = Model()
my_model.add_instantiated_nodes([my_node, my_dist])
my_model.add_instantiated_arcs([my_arc])

# Inspect push_check_handler for distribution
print(my_model.nodes["my_dist"].push_check_handler)

# %% [markdown]
# We have created a `Node` that is connected to a `Distribution`, and we can see by inspecting its `push_check_handler`
# that the `Distribution` object has a denial behaviour for push checks (i.e., it says it cannot be pushed to). We can
# verify this by sending a `push_check`.
#
# Remember that a `push_check` requires sending a VQIP (a dictionary with a key for 'volume' and each pollutant
# simulated).
# %%
reply = my_model.arcs["my_arc"].send_push_check(
    {"volume": 10, "phosphate": 1, "temperature": 10}
)
print(reply)


# %% [markdown]
# The `Distribution` node replied that it can accept 0 water.
#
# But we might try and customise the `my_dist` object so that it calls some function before carrying on with its default push check.
#
# Let's have a look at what code we need to store in our extension module, which we have saved in a file called `custom_distribution_handler.py`.
# We will reproduce the function below, which will cause a warning when we load our model because the patch is registered twice.
# %%
from wsimod.extensions import register_node_patch


@register_node_patch("my_dist", "push_check_handler", item="Node")
def custom_handler_function(self, vqip, *args, **kwargs):
    """A custom `push_check_handler` function.

    Call the default handler for the "Node" item.
    """
    print("I reached a custom handler")
    return self.push_check_handler["default"](vqip)


# %% [markdown]
# We can customise our model with this handler by specifying it under the
# `extensions` attribute and reloading the model to apply the extension.
# %%
my_model.extensions = [os.path.join(scripts_folder, "custom_distribution_handler.py")]

my_model.save(temp_dir.name)
my_model.load(temp_dir.name)

print(my_model.nodes["my_dist"].push_check_handler)

# %% [markdown]
# We have added a new custom function for the `Node` tag! It includes a print statement so we should be able to see if it is triggered.
#
# Let's see what happens when we call that node via it's connected arc. Note, this is an example of how custom handlers **do not** work!
# %%
reply = my_model.arcs["my_arc"].send_push_check(
    {"volume": 10, "phosphate": 1, "temperature": 10}
)
print(reply)

# %% [markdown]
# Even though `my_arc` starts at a `Node` object, the custom handler wasn't used. That is because, as explained earlier, if the tag is not specified, the `default` tag will always be used, no matter what node type the push check is originating from. Let us instead specify a tag. Note, this is an example of how custom handlers **do** work!

# %%
reply = my_model.arcs["my_arc"].send_push_check(
    {"volume": 10, "phosphate": 1, "temperature": 10}, tag="Node"
)
print(reply)

# %% [markdown]
# Great - we have successfully customised a handler and used it properly.

# %% [markdown]
# ## More realistic example
# A typical kind of behaviour we might be keen to introduce through handlers would be customising how the `Reservoir` object
# responds to pulls. By default, any node can pull from a reservoir:

# %%
from wsimod.nodes.storage import Reservoir

my_reservoir = Reservoir(name="my_reservoir", capacity=10, initial_storage=9)
print(my_reservoir.pull_set_handler)

# %% [markdown]
# We see in the `pull_set_handler` that all pull requests are channeled through the `default` tag function, which presumably
# updates the reservoir storage. Let's just verify that:

# %%
# Inspect initial conditions
print("Initial storage: {0}".format(my_reservoir.tank.storage))

# Send a pull request
reply = my_reservoir.pull_set_handler["default"]({"volume": 1})

# Inspect new conditions
print("Amount pulled: {0}".format(reply))
print("Remaining storage: {0}".format(my_reservoir.tank.storage))

# %% [markdown]
# Behaviour as expected, but this might be a problem. Especially if this reservoir is intended only as a raw water supply,
# we may want to customise its handlers such that only the `FWTW` object can pull from it.
#
# Well if we want to do this we will have to take some steps to implement this:
#  1. Create a simple model to test this behaviour.
#  2. Update the reservoir's `pull_set_handler` and `pull_check_handler`.
#  3. Ensure that the `FWTW` pulls from a `Reservoir` with the `FWTW` tag.

# %% [markdown]
# ### 1. Create a test model
# We will create two nodes that pull water, a [`FWTW`](https://imperialcollegelondon.github.io/wsi/reference-wtw/#wsimod.nodes.wtw.FWTW)
# and a [`Demand`](https://imperialcollegelondon.github.io/wsi/reference-other/#wsimod.nodes.demand.Demand). And link them both to
# `my_reservoir`. Under the default behaviour of a `Reservoir`, it can be pulled from both of these node types, but we may decide
# that the water is not clean enough to go straight to `Demand`, and thus wish to customise the handler. (Of course we could more
# simply remove the arc between `my_demand` and `my_reservoir` - but you will have to use your imagination and decide that it is
# better to customise the handler).

# %%
# We will use the previous reservoir

# Create a FWTW
from wsimod.nodes.wtw import FWTW

my_fwtw = FWTW(
    name="my_fwtw",
    service_reservoir_storage_capacity=2,
)

# Create another object to pull from the reservoir, e.g., a demand node
from wsimod.nodes.demand import Demand

my_demand = Demand(name="my_demand", constant_demand=1.5)

# Link both objects to the reservoir
reservoir_to_fwtw = Arc(
    name="reservoir_to_fwtw", in_port=my_reservoir, out_port=my_fwtw
)
reservoir_to_demand = Arc(
    name="reservoir_to_demand", in_port=my_reservoir, out_port=my_demand
)

# Store everything to a model
my_model = Model()
my_model.add_instantiated_nodes([my_reservoir, my_fwtw, my_demand])
my_model.add_instantiated_arcs([reservoir_to_fwtw, reservoir_to_demand])

# %% [markdown]
# By inspecting their documentation, we see that the
# [`create_demand`](https://imperialcollegelondon.github.io/wsi/reference-other/#wsimod.nodes.demand.Demand.create_demand)
# function pulls water for a `Demand` node, while the
# [`treat_water`](https://imperialcollegelondon.github.io/wsi/reference-wtw/#wsimod.nodes.wtw.FWTW.treat_water)
# does so for a `FWTW` node.

# %%
# Inspect initial conditions
print("Initial storage: {0}".format(my_reservoir.tank.storage))
print(
    "Initial service reservoir storage: {0}".format(
        my_fwtw.service_reservoir_tank.storage
    )
)

# Run model for a timestep
results = my_model.run(dates=[to_datetime("2000-01-01")])

# Inspect new conditions
print("Remaining reservoir storage: {0}".format(my_reservoir.tank.storage))
print(
    "New service reservoir storage: {0}".format(my_fwtw.service_reservoir_tank.storage)
)

# View arc flows
print(results[0])

# %% [markdown]
# We see some complaint from the `my_fwtw` about sludge (which are also explaining the small losses and why the new service reservoir volume is not quite 2), because it is anticipated to be connected to a sewer, but by inspecting the reservoir volumes we see that the `FWTW` has correctly pulled from the `Reservoir`.
#
# Again, a complaint that `my_demand` had nowhere to send its sewage, but we can see that it has successfully pulled
# water from `my_reservoir`. Now we will update `my_reservoir` so that it denies pulls unless they have the tag `FWTW`.

# %% [markdown]
# ### 2. Update reservoir handlers
# As we illustrated [above](#illustration), we define our new handlers in a separate module, which we will call `custom_reservoir_handler.py`.


# %%
@register_node_patch("my_reservoir", "pull_set_handler", item="FWTW")
def custom_pulls_fwtw(self, vqip, *args, **kwargs):
    """A custom `pull_set_handler` function.

    Pull from the storage when pulled with the tag "FWTW".
    """
    return self.tank.pull_storage(vqip)


@register_node_patch("my_reservoir", "pull_check_handler", item="FWTW")
def custom_pullc_fwtw(self, vqip, *args, **kwargs):
    """A custom `pull_check_handler` function.

    Return available storage when pulled with the tag "FWTW".
    """
    return self.tank.get_avail()


@register_node_patch("my_reservoir", "pull_set_handler", item="default")
def custom_pulls_default(self, vqip, *args, **kwargs):
    """A custom `pull_set_handler` function.

    Deny pull sets by default.
    """
    return self.pull_set_deny(vqip)


@register_node_patch("my_reservoir", "pull_check_handler", item="default")
def custom_pullc_default(self, vqip, *args, **kwargs):
    """A custom `pull_check_handler` function.

    Deny pull checks by default.
    """
    return self.pull_check_deny()


# %% [markdown]
# Lets add this extension to the model and reload it to apply the extension, verifying that the handler functions have changed.
# %%
# Inspect original handlers
print(
    "Original set handler: {0}".format(my_model.nodes["my_reservoir"].pull_set_handler)
)
print(
    "Original check handler: {0}".format(
        my_model.nodes["my_reservoir"].pull_check_handler
    )
)

# Clear the extensions registry (otherwise it will try and apply the earlier extension to `my_dist`)
from wsimod.extensions import extensions_registry

extensions_registry.clear()

# Reload to apply the extensions
my_model.extensions = [os.path.join(scripts_folder, "custom_reservoir_handler.py")]
my_model.save(temp_dir.name)
my_model.load(temp_dir.name)

# Inspect new handlers
print(
    "Overwritten set handler: {0}".format(
        my_model.nodes["my_reservoir"].pull_set_handler
    )
)
print(
    "Overwritten set handler: {0}".format(
        my_model.nodes["my_reservoir"].pull_check_handler
    )
)

# %% [markdown]
# It appears that we have successfully overwritten the handlers.
#
# Let's verify this behaviour by ensuring that `my_demand` cannot pull from the reservoir
# %%
# Inspect initial conditions
print("Initial storage: {0}".format(my_model.nodes["my_reservoir"].tank.storage))
print(
    f"Initial storage in service reservoirs: {my_model.nodes['my_fwtw'].service_reservoir_tank.storage}"
)

# Call function to trigger pulls
results = my_model.run(dates=[to_datetime("2000-01-01")])

# Inspect new conditions
print(
    "Remaining reservoir storage: {0}".format(
        my_model.nodes["my_reservoir"].tank.storage
    )
)
print(
    f"New storage in service reservoirs: {my_model.nodes['my_fwtw'].service_reservoir_tank.storage}"
)

# View arc flows
print(results[0])

# %% [markdown]
# Fantastic - now we see that `my_demand` is not able to pull from `my_reservoir`, triggering a bunch of warning
# messages that occur when trying to pull from a node with denial behaviour.
#
# However, we also find that now `my_fwtw` cannot pull from `my_reservoir`! As we have explained [above](#more-realistic-example), this is because when `my_fwtw` makes a pull, it sends the `default` tag by default, not the `FWTW` tag.
#
# Note that we see that the service_reservoir volume has increased, this is because
# of the [assumption](https://imperialcollegelondon.github.io/wsi/reference-wtw/#wsimod.nodes.wtw.FWTW) that FWTW deficits will be met by other measures
#
# %% [markdown]
# ### 3. Update `my_fwtw` to pull with `FWTW` tag
# Now we need to ensure that `my_fwtw` includes the `FWTW` tag when it pulls from the reservoir.
# By inspecting the documentation, we see that `FWTW` pulls water in the
# [`treat_water`](https://imperialcollegelondon.github.io/wsi/reference-wtw/#wsimod.nodes.wtw.FWTW.treat_water)
# function using the
# [`pull_distributed`](https://imperialcollegelondon.github.io/wsi/reference-nodes/#wsimod.nodes.nodes.Node.pull_distributed)
# function.
#
# Thus, we could either overwrite the `treat_water` function or the `pull_distributed` function.
#
# If we choose the first option, we will essentially have to rewrite the whole function, but ensure that
# `pull_distributed` is called with the tag `FWTW`.
#
# If we chose the second option, we can just use a decorator to ensure that every time the `pull_distributed`
# function is called, it is called with the tag `FWTW`. Since this seems simpler, we will choose this option.
#
# We have the following code in a separate module, which we will call `custom_fwtw_pull.py`.
# %%


@register_node_patch("my_fwtw", "pull_distributed")
def custom_pull_distributed(self, vqip, *args, **kwargs):
    """A custom `pull_distributed` function.

    Call `pull_distributed` with the tag "FWTW".
    """
    return self._patched_pull_distributed(vqip, tag="FWTW")


# %% [markdown]
# Again, we will add this extension to the model and reload it to apply the extension, verifying that the handler function has been applied and is working.
# %%
# Add the extension to the model
my_model.extensions.append(os.path.join(scripts_folder, "custom_fwtw_pull.py"))

# Reload to apply the extensions
extensions_registry.clear()
my_model.save(temp_dir.name)
my_model.load(temp_dir.name)

# Run the model
results = my_model.run(dates=[to_datetime("2000-01-01")])

# Inspect new conditions
print(
    "Remaining reservoir storage: {0}".format(
        my_model.nodes["my_reservoir"].tank.storage
    )
)
print(
    f"New storage in service reservoirs: {my_model.nodes['my_fwtw'].service_reservoir_tank.storage}"
)

# View arc flows
print(results[0])

# %% [markdown]
# Fantastic, we've got the results we wanted and have now customised a handler!

# %%
# Close the temporary directory
temp_dir.cleanup()

# %% [markdown]
# ## What next?
# Surely you are an expert at WSIMOD by now! Why not check our
# [contribution guidelines](https://github.com/imperialcollegelondon/wsi/blob/main/docs/CONTRIBUTING.md)!



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.demand import Demand, ResidentialDemand
from wsimod.nodes.land import Land
from wsimod.nodes.waste import Waste
from wsimod.orchestration.model import to_datetime


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_get_demand(self):
        demand = Demand(
            name="",
            constant_demand=10,
            pollutant_load={"phosphate": 0.1, "temperature": 12},
        )

        reply = demand.get_demand()
        d1 = {"volume": 10, "phosphate": 0.1, "temperature": 12}
        self.assertDictAlmostEqual(d1, reply["default"])

    def test_create_demand(self):
        demand = Demand(
            name="",
            constant_demand=10,
            pollutant_load={"phosphate": 0.1, "temperature": 12},
        )

        waste = Waste(name="")
        arc1 = Arc(in_port=demand, out_port=waste, name="")
        demand.create_demand()
        d1 = {"volume": 10, "phosphate": 0.1, "temperature": 12}
        self.assertDictAlmostEqual(d1, arc1.vqip_out)
        self.assertDictAlmostEqual(d1, demand.total_demand)

        d2 = {"volume": 0, "phosphate": 0, "temperature": 0}
        self.assertDictAlmostEqual(d2, demand.total_received)

    def test_house_demand(self):
        demand = ResidentialDemand(
            name="",
            pollutant_load={"phosphate": 0.1},
            data_input_dict={("temperature", 1): 10},
            population=5,
            per_capita=0.12,
            constant_weighting=0.4,
            constant_temp=8,
        )
        demand.t = 1
        reply = demand.get_house_demand()
        d1 = {
            "phosphate": 5 * 0.1,
            "volume": 5 * 0.12,
            "temperature": 10 * 0.6 + 8 * 0.4,
        }

        self.assertDictAlmostEqual(d1, reply)

    def test_excess_to_garden(self):
        demand = ResidentialDemand(name="", gardening_efficiency=0.3)

        self.assertEqual(0.36, demand.excess_to_garden_demand(1.2))

    def test_garden_demand(self):
        date = to_datetime("2000-05-01")
        demand = ResidentialDemand(name="", gardening_efficiency=0.4)
        demand.t = date

        land = Land(
            name="",
            surfaces=[
                {
                    "type_": "GardenSurface",
                    "area": 1,
                    "rooting_depth": 1,
                    "field_capacity": 0.3,
                    "wilting_point": 0.1,
                    "initial_storage": 0.2,
                }
            ],
            data_input_dict={
                ("temperature", date): 15,
                ("precipitation", date): 0,
                ("et0", date): 0.03,
            },
        )
        land.t = date
        Arc(in_port=demand, out_port=land, name="")
        land.run()
        reply = demand.get_garden_demand()

        d1 = {"volume": 0.03 * 0.4, "temperature": 0, "phosphate": 0}
        self.assertDictAlmostEqual(d1, reply)

    def test_demand_overrides(self):
        demand = Demand(
            name="",
            constant_demand=10,
            pollutant_load={"phosphate": 0.1, "temperature": 12},
        )
        demand.apply_overrides(
            {"constant_demand": 20, "pollutant_load": {"phosphate": 0.5}}
        )
        self.assertEqual(demand.constant_demand, 20)
        self.assertDictEqual(
            demand.pollutant_load, {"phosphate": 0.5, "temperature": 12}
        )

    def test_residentialdemand_overrides(self):
        demand = ResidentialDemand(
            name="",
            gardening_efficiency=0.4,
            pollutant_load={"phosphate": 0.1, "temperature": 12},
        )
        demand.apply_overrides(
            {
                "gardening_efficiency": 0.5,
                "population": 153.2,
                "per_capita": 32.4,
                "constant_weighting": 47.5,
                "constant_temp": 0.71,
                "constant_demand": 20,
                "pollutant_load": {"phosphate": 0.5},
            }
        )
        self.assertEqual(demand.gardening_efficiency, 0.5)
        self.assertEqual(demand.population, 153.2)
        self.assertEqual(demand.per_capita, 32.4)
        self.assertEqual(demand.constant_weighting, 47.5)
        self.assertEqual(demand.constant_temp, 0.71)

        self.assertEqual(demand.constant_demand, 20)
        self.assertDictEqual(
            demand.pollutant_load, {"phosphate": 0.5, "temperature": 12}
        )


if __name__ == "__main__":
    unittest.main()



# %% [markdown]
# # Land nodes - hydrology and agriculture (.py)
#
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/demo/scripts](https://github.com/barneydobson/wsi/blob/main/docs/demo/scripts/land_demo.py)
#
# 1. [Introduction](#introduction)
#
# 2. [Data](#imports-and-forcing-data)
#
# 3. [Basic surface](#basic-surface)
#
# 4. [Pervious surface](#pervious-surface)
#
# 5. [Connecting land nodes in a model](#connecting-land-nodes-in-a-model)
#
#     5.1 [Model object](#model-object)
#
# 6. [Growing surface](#growing-surface)
#
# %% [markdown]
# ## Introduction
#
# A land node is the object that interfaces with other WSIMOD nodes.
# Each land node can have one or more surfaces.
# Each surface can be parameterised differently or follow different equations,
# similar to the hyrological response unit concept in hydrological models.
#
# The different surfaces within WSIMOD are used to capture hydrological,
# agricultural, and urban runoff processes. In this demo we will see how they
# all work.
#
# %% [markdown]
# ## Imports and forcing data

# %% [markdown]
# Import packages
# %%

import os

import pandas as pd
from matplotlib import pyplot as plt

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.demo.create_oxford import create_timeseries
from wsimod.nodes.land import Land
from wsimod.nodes.nodes import Node
from wsimod.nodes.storage import Groundwater
from wsimod.nodes.waste import Waste
from wsimod.orchestration.model import Model

# %% [markdown]
# Load input data
# %%

# Select the root path for the data folder. Use the appropriate value for your case.
data_folder = os.path.join(os.path.abspath(""), "docs", "demo", "data")

input_fid = os.path.join(data_folder, "processed", "timeseries_data.csv")
input_data = pd.read_csv(input_fid)
input_data.loc[input_data.variable == "flow", "value"] *= constants.M3_S_TO_M3_DT
input_data.loc[input_data.variable == "precipitation", "value"] *= constants.MM_TO_M
input_data.date = pd.to_datetime(input_data.date)
data_input_dict = input_data.set_index(["variable", "date"]).value.to_dict()
data_input_dict = (
    input_data.groupby("site")
    .apply(lambda x: x.set_index(["variable", "date"]).value.to_dict())
    .to_dict()
)

dates = input_data.date.drop_duplicates()
dates_monthyear = input_data.date.dt.to_period("M").unique()
print(input_data.sample(10))

# %% [markdown]
# Input data is stored in dicts
# %%

land_inputs = data_input_dict["oxford_land"]

example_date = pd.to_datetime("2009-03-03")

print(land_inputs[("precipitation", example_date)])
print(land_inputs[("et0", example_date)])
print(land_inputs[("temperature", example_date)])


# %% [markdown]
# We are just using some basic pollutants for demonstration
# %%

constants.set_simple_pollutants()

print(constants.POLLUTANTS)


# %% [markdown]
# ## Basic surface

# %% [markdown]
# Create a simple [Land node](./../../../reference-land/#wsimod.nodes.land.Land)
# with one [basic surface](./../../../reference-land/#wsimod.nodes.land.Surface).
# We can pass surfaces as a dictionary when creating a Land node.
# %%
surface = {
    "type_": "Surface",
    "surface": "my_surface",
    "area": 10,
    "depth": 1,
    "pollutant_load": {"phosphate": 1e-7},
}

land = Land(name="my_land", data_input_dict=land_inputs, surfaces=[surface])

# %% [markdown]
# Each surface object is stored in a list of surfaces within the land node.
# %%
print(land.surfaces[0])

# %% [markdown]
# We can also access a specific surface by name.
# %%
print(land.get_surface("my_surface"))

# %% [markdown]
# This surface will have all the data we passed to the land node in a dict.
# %%
print(land.get_surface("my_surface").area)
print(land.get_surface("my_surface").pollutant_load)

# %% [markdown]
# A surface is a generic tank, by default it is initialised as empty.
# %%
print(land.get_surface("my_surface").storage)


# %% [markdown]
# Let's try to run a timestep for a land node
# %%
land.t = example_date
land.monthyear = land.t.to_period("M")
land.run()


# %% [markdown]
# We see that a day of phosphate deposition has occured on the surface.
# %%
print(land.get_surface("my_surface").storage)

# %% [markdown]
# The run function in land simply calls the run function in all of its surfaces.
# %%

land.get_surface("my_surface").run()

# %% [markdown]
# After running the surface, we see that another day of phosphate deposition
# has occured.
# %%

print(land.get_surface("my_surface").storage)


# %% [markdown]
# However... no precipitation has occurred (no volume above),
# despite there being rainfall!
# %%
print(land.get_data_input("precipitation"))

# %% [markdown]
# To understand why this is, we have to understand how the 'run' function in a
# surface works. Each surface has a list of functions stored in 'inflows', in
# 'processes', and in 'outflows'. When 'run' is called, all of the functions in
# inflows are executed, then processes, then outflows.
#
# If we print the lists of functions, we see only some simple deposition occurs
# on the basic surface in the inflows...
# %%

print(land.get_surface("my_surface").inflows)
print(land.get_surface("my_surface").processes)
print(land.get_surface("my_surface").outflows)

# %% [markdown]
# We need a specific subclass of surface to implement more sophisticated
# hydrological processes!
#
# %% [markdown]
# ## Pervious surface

# %% [markdown]
# The pervious surface contains a simple lumped hydrological model.
#
# We can create the surface, again via parameters stored in a dictionary that
# is passed to the Land node.
# is passed to the Land node. Note that the subclass of surface to be created
# must be specified by the `type_` keyword.
# %%
surface = {
    "type_": "PerviousSurface",
    "surface": "my_surface",
    "area": 10,
    "depth": 0.5,
    "pollutant_load": {"phosphate": 1e-7},
    "wilting_point": 0.05,
    "field_capacity": 0.1,
}

land = Land(name="my_land", data_input_dict=land_inputs, surfaces=[surface])

# %% [markdown]
# We have lots of functions now! You will have to look at the documentation of
# [PerviousSurface](./../../../reference-land/#wsimod.nodes.land.PerviousSurface)
# to understand them in detail
#
# We note that the function 'ihacres' in the inflows is the hydrological
# processes representation, using equations based off the IHACRES model.
# If a user preferred to use a different hydrological model, they would simply
# need to substitute out this function in the inflows list.
# %%

print(land.get_surface("my_surface").inflows)
print(land.get_surface("my_surface").processes)
print(land.get_surface("my_surface").outflows)

# %% [markdown]
# Now when we run the model, we can see that some rain has happened, because the
# storage volume has increased (as well as the deposition).
# %%
land.t = example_date
land.monthyear = land.t.to_period("M")
land.run()
print(land.get_surface("my_surface").storage)

# %% [markdown]
# However, if we look at the land node, which is the parent of the surfaces that
# interfaces with other WSIMOD components, we see that all of its tanks are still
# empty.
# %%
print(land.percolation.storage)
print(land.subsurface_runoff.storage)
print(land.surface_runoff.storage)

# %% [markdown]
# This is because the IHACRES model requires soil moisture to be more than a
# specified amount (the field capacity) for flows to be generated.
#
# We can run the surface multiple times to fill it up with water.
# %%
for i in range(10):
    land.get_surface("my_surface").run()

# %% [markdown]
# There's now a lot of water in the tank
# %%
print(land.get_surface("my_surface").storage)


# %% [markdown]
# Critically, we see that the moisture content is greater than 0.1
# (i.e., the field capacity moisture content)
# %%
print(land.get_surface("my_surface").get_smc() / land.get_surface("my_surface").depth)

# %% [markdown]
# Once soil moisture content is greater than the field capacity,
# flows will be generated and the land tanks will fill up.
# %%
print(land.percolation.storage)
print(land.subsurface_runoff.storage)
print(land.surface_runoff.storage)


# %% [markdown]
# These tanks represent flow from the soil layer to either rivers or groundwater.
# However, land nodes expect to be able to route flow onwards to other nodes.
#
# Since the land isn't connected to anything, these won't actually go anywhere
# if we run it, and they will just build up
# %%
for i in range(10):
    land.run()

print(land.percolation.storage)
print(land.subsurface_runoff.storage)
print(land.surface_runoff.storage)

# %% [markdown]
# ## Connecting land nodes in a model
# As mentioned, the land node expects to be able to discharge to [groundwater](./../../../reference-storage/#wsimod.nodes.storage.Groundwater),
# and rivers (where a river could be the
# [River node](./../../../reference-storage/#wsimod.nodes.storage.River)
# or just a generic [Node](./../../../reference-nodes/#wsimod.nodes.nodes.Node)).
# We also provide a [Waste node](./../../../reference-other/#wsimod.nodes.waste.Waste),
# which is just a model outlet.
# %%
node = Node(name="my_river")
gw = Groundwater(name="my_groundwater", area=10, capacity=100)
outlet = Waste(name="my_outlet")

# %% [markdown]
# We use [arcs](./../../../reference-arc/#wsimod.arcs.arcs.Arc) to join up all of the different nodes according to a standard
# hydrological representation.
# %%
arc1 = Arc(in_port=land, out_port=node, name="quickflow")
arc2 = Arc(in_port=land, out_port=gw, name="percolation")
arc3 = Arc(in_port=gw, out_port=node, name="baseflow")
arc4 = Arc(in_port=node, out_port=outlet, name="outflow")

# %% [markdown]
# If we run the land a few more times, we see that these tanks start to empty
# (though percolation by nature empties rather slowly!!!)
# %%

for i in range(10):
    land.run()

print(land.percolation.storage)
print(land.subsurface_runoff.storage)
print(land.surface_runoff.storage)

# %% [markdown]
# ### Model object
# We can put these nodes and arcs into the [Model object](./../../../reference-model/#wsimod.orchestration.model.Model)
# to have a functioning hydrological model.
#
# We start by creating a model object.
# %%

my_model = Model()

# %% [markdown]
# Since we have already created our nodes/arcs, we use the add_instantiated functions
# %%
my_model.add_instantiated_nodes([land, node, gw, outlet])
my_model.add_instantiated_arcs([arc1, arc2, arc3, arc4])

# %% [markdown]
# Store dates
# %%
my_model.dates = dates

# %% [markdown]
# We have run the surfaces a few times, so will just set all of the model tanks
# to empty to give a clean start for the model.
# %%
my_model.reinit()

# %% [markdown]
# We can run the model with the run function
# %%
results = my_model.run()

# %% [markdown]
# .. and plot the results!
# %%
flows = pd.DataFrame(results[0])

f, axs = plt.subplots(2, 1)
flows.groupby("arc").get_group("outflow").set_index("time").flow.plot(ax=axs[0])
flows.groupby("arc").get_group("outflow").set_index("time").phosphate.plot(ax=axs[1])

# %% [markdown]
# ## Growing surface
# Hydrology is nice, but anyone using WSIMOD probably isn't interested in
# hydrology only! The GrowingSurface adds a lot of sophisticated behaviour for
# agriculture and water quality.
#
# Our GrowingSurface needs a bit more data than other surfaces, for fertiliser,
# manure and atmospheric deposition of ammonia, nitrate and phosphate.
# We will make up this data.
#
# Surface pollution data varies at a monthly timestep rather than daily,
# though it is applied each day.
# %%

surface_input_data = {}
for pollutant in ["srp", "nhx", "noy"]:
    for source in ["manure", "fertiliser", "dry", "wet"]:
        amount = 1e-7  # kg/m2/timestep
        ts = create_timeseries(
            amount, dates_monthyear, "{0}-{1}".format(pollutant, source)
        )
        ts = ts.set_index(["variable", "date"]).value.to_dict()
        surface_input_data = {**surface_input_data, **ts}

print(surface_input_data[("nhx-manure", example_date.to_period("M"))])

# %% [markdown]
# Because the surface represents the nitrogen and phosphorus cycles, we need
# to simulate a greater number of pollutants
# %%
constants.set_default_pollutants()

# %% [markdown]
# As with the other surfaces, we create the surface by passing it as a dictionary
# to the created Land node. I will use the parameters for Maize for this growing surface
# %%
crop_factor_stages = [0.0, 0.0, 0.3, 0.3, 1.2, 1.2, 0.325, 0.0, 0.0]
crop_factor_stage_dates = [0, 90, 91, 121, 161, 213, 244, 245, 366]
sowing_day = 91
harvest_day = 244
ET_depletion_factor = 0.55
rooting_depth = 0.5

surface = {
    "type_": "GrowingSurface",
    "surface": "my_growing_surface",
    "area": 10,
    "rooting_depth": rooting_depth,
    "crop_factor_stage_dates": crop_factor_stage_dates,
    "crop_factor_stages": crop_factor_stages,
    "sowing_day": sowing_day,
    "harvest_day": harvest_day,
    "ET_depletion_factor": ET_depletion_factor,
    "data_input_dict": surface_input_data,
    "wilting_point": 0.05,
    "field_capacity": 0.1,
}

land = Land(name="my_land", data_input_dict=land_inputs, surfaces=[surface])

# %% [markdown]
# We see that the inflows includes IHACRES from the pervious surface.
# But has also added a range of other functions related to deposition.
# %%
print(land.get_surface("my_growing_surface").inflows)

# %% [markdown]
# And in particular with the bio-chemical processes occurring within nutrient pools
# %%
print(land.get_surface("my_growing_surface").processes)

# %% [markdown]
# Let's recreate our model with this Maize surface and run the model
# %%
node = Node(name="my_river")
gw = Groundwater(name="my_groundwater", area=10, capacity=100)
outlet = Waste(name="my_outlet")

arc1 = Arc(in_port=land, out_port=node, name="quickflow")
arc2 = Arc(in_port=land, out_port=gw, name="percolation")
arc3 = Arc(in_port=gw, out_port=node, name="baseflow")
arc4 = Arc(in_port=node, out_port=outlet, name="outflow")
my_model = Model()

my_model.add_instantiated_nodes([land, node, gw, outlet])
my_model.add_instantiated_arcs([arc1, arc2, arc3, arc4])

my_model.dates = dates

results = my_model.run()

# %% [markdown]
# We can now plot the results
# %%

flows = pd.DataFrame(results[0])

f, axs = plt.subplots(2, 1)
flows.groupby("arc").get_group("outflow").set_index("time").flow.plot(ax=axs[0])
flows.groupby("arc").get_group("outflow").set_index("time").phosphate.plot(ax=axs[1])

# %% [markdown]
# Observe the differences between the two sets of timeseries:
# Flows look more or less the same (dynamically), which makes sense since they
# both use IHACRES for hydrology. Only small differences will arise because
# the crops change the evapotranspiration coefficient
#
# Meanwhile, phosphate levels look much more interesting with the
# GrowingSurface, and are not solely dependent on the hydrology.
